"""Script to run a jupyter notebook for a particular model."""

import argparse
import logging
import os
import sys
import tempfile

import papermill as pm

import flyvis
from flyvis.utils.config_utils import HybridArgumentParser

if __name__ == "__main__":
    parser = HybridArgumentParser(
        hybrid_args={
            "notebook_per_model": {
                "type": bool,
                "required": False,
                "help": (
                    "Flag to set the output path to the per-model notebook path. "
                    "Requires ensemble_and_network_id."
                ),
            },
            "ensemble_and_network_id": {
                "type": str,
                "required": False,
                "help": (
                    "Id in form of task_name/ensemble_id/network_id, e.g., "
                    "flow/0000/000. Required if notebook_per_model is True."
                ),
            },
            "per_ensemble": {
                "type": bool,
                "required": False,
                "help": (
                    "Flag to set the output path to the per-ensemble notebook path. "
                    "Requires ensemble_id and task_name."
                ),
            },
            "task_name": {
                "type": str,
                "required": False,
                "help": (
                    "Name of the task, e.g. flow. Required if per_ensemble is True."
                ),
            },
            "ensemble_id": {
                "type": int,
                "required": False,
                "help": (
                    "Id of the ensemble, e.g. 0045. Required if per_ensemble is True."
                ),
            },
        },
        description=(
            "Run a Jupyter notebook using papermill. "
            "Required arguments depend on the specific notebook. "
            "Pass any additional arguments as key:type=value triplets."
        ),
        usage=(
            "\nflyvis notebook [-h] --notebook_path PATH [options] [hybrid_args...]\n"
            "       or\n"
            "%(prog)s [-h] --notebook_path PATH [options] [hybrid_args...]\n"
            "\n"
            "There are three modes of operation:\n"
            "1. Basic: Only requires --notebook_path\n"
            "2. Per-model: Requires --notebook_per_model_path, "
            "--notebook_per_model=true, --ensemble_and_network_id\n"
            "3. Per-ensemble: Requires --notebook_path, --per_ensemble=true, "
            "--task_name, --ensemble_id\n"
        ),
        epilog="""
        Examples:
            # Basic usage
            flyvis notebook --notebook_path notebooks/analysis.ipynb
            # Per-model analysis
            flyvis notebook --notebook_per_model_path notebooks/per_model.ipynb \\
                            notebook_per_model=true \\
                            ensemble_and_network_id=flow/0045/000
            # Per-ensemble analysis
            flyvis notebook --notebook_path notebooks/ensemble.ipynb \\
                            per_ensemble=true \\
                            task_name=flow \\
                            ensemble_id=45
            # With additional parameters passed to the notebook
            flyvis notebook --notebook_path notebooks/analysis.ipynb \\
                            param1:str=value1 \\
                            param2:int=42
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--notebook_path",
        type=str,
        help="Path of the notebook to execute, e.g. /path/to/__main__.ipynb.",
    )
    parser.add_argument(
        "--notebook_per_model_path",
        type=str,
        help=(
            "Path of the notebook to execute for each model, e.g. "
            "/path/to/__main_per_model__.ipynb."
        ),
    )
    parser.add_argument(
        "--output_path",
        type=str,
        default=None,
        help=(
            "Path for the output notebook. If not provided, a temporary file "
            "will be used and deleted after execution."
        ),
    )
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Perform a dry run without actually executing the notebook.",
    )

    # arguments are parsed including hybrid and typed hybrid arguments and directly
    # passed to papermill
    args = parser.parse_with_hybrid_args()

    if args.dry:
        logging.info(
            "Dry run. Not executing notebook %s. Passed args: %s",
            args.notebook_path,
            args,
        )
        sys.exit(0)

    logging.info("Executing notebook %s", args.notebook_path)
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as f:
        if args.notebook_per_model:
            input_path = args.notebook_per_model_path
            output_path = (
                flyvis.results_dir
                / f"{args.task_name}/{args.ensemble_and_network_id}"
                / args.notebook_per_model_path.split(os.sep)[-1]
            )

        elif args.per_ensemble:
            input_path = args.notebook_path
            output_path = (
                flyvis.results_dir
                / f"{args.task_name}/{args.ensemble_id:04}"
                / args.notebook_path.split(os.sep)[-1]
            )
        else:
            output_path = args.output_path or f.name
        logging.info("Output will be saved to %s.", output_path)
        pm.execute_notebook(
            flyvis.repo_dir / input_path,
            output_path,
            parameters=args.__dict__,
        )
    logging.info("Done executing notebook %s", input_path)



"""Utility function for operations on hexagonal lattices."""

from numbers import Number
from typing import Iterable, Literal, Tuple

import numpy as np
import torch
from matplotlib import colormaps as cm
from numpy.typing import NDArray

import flyvis
from flyvis.utils.tensor_utils import matrix_mask_by_sub


def get_hex_coords(extent: int, astensor: bool = False) -> Tuple[NDArray, NDArray]:
    """Construct hexagonal coordinates for a regular hex-lattice with extent.

    Args:
        extent: Integer radius of hexagonal lattice. 0 returns the single
            center coordinate.
        astensor: If True, returns torch.Tensor, else np.array.

    Returns:
        A tuple containing:
            u: Hex-coordinates in u-direction.
            v: Hex-coordinates in v-direction.

    Note:
        Will return `get_num_hexals(extent)` coordinates.

    See Also:
        https://www.redblobgames.com/grids/hexagons/#range-coordinate
    """
    u = []
    v = []
    for q in range(-extent, extent + 1):
        for r in range(max(-extent, -extent - q), min(extent, extent - q) + 1):
            u.append(q)
            v.append(r)
    if astensor:
        return torch.tensor(u, dtype=torch.long), torch.tensor(v, dtype=torch.long)
    return np.array(u), np.array(v)


def hex_to_pixel(
    u: NDArray,
    v: NDArray,
    size: float = 1,
    mode: Literal["default", "flat", "pointy"] = "default",
) -> Tuple[NDArray, NDArray]:
    """Returns pixel coordinates from hex coordinates.

    Args:
        u: Hex-coordinates in u-direction.
        v: Hex-coordinates in v-direction.
        size: Size of hexagon.
        mode: Coordinate system convention.

    Returns:
        A tuple containing:
            x: Pixel-coordinates in x-direction.
            y: Pixel-coordinates in y-direction.

    See Also:
        https://www.redblobgames.com/grids/hexagons/#hex-to-pixel
    """
    if isinstance(u, list) and isinstance(v, list):
        u = np.array(u)
        v = np.array(v)
    if mode == "default":
        return 3 / 2 * v, -np.sqrt(3) * (u + v / 2)
    elif mode == "flat":
        return (3 / 2 * u) * size, (np.sqrt(3) / 2 * u + np.sqrt(3) * v) * size
    elif mode == "pointy":
        return (np.sqrt(3) * u + np.sqrt(3) / 2 * v) * size, (3 / 2 * v) * size
    else:
        raise ValueError(f"{mode} not recognized.")


def hex_rows(
    n_rows: int,
    n_columns: int,
    eps: float = 0.1,
    mode: Literal["pointy", "flat"] = "pointy",
) -> Tuple[NDArray, NDArray]:
    """Return a hex grid in pixel coordinates.

    Args:
        n_rows: Number of rows.
        n_columns: Number of columns.
        eps: Small offset to avoid overlapping hexagons.
        mode: Orientation of hexagons.

    Returns:
        A tuple containing:
            x: X-coordinates of hexagon centers.
            y: Y-coordinates of hexagon centers.
    """
    u = []
    v = []
    for r in range(n_rows):
        for c in range(n_columns):
            u.append(c)
            v.append(r)
    u = np.array(u)
    v = np.array(v)
    x, y = hex_to_pixel(u, v, mode=mode)
    x += eps
    y += eps
    return x, y


def pixel_to_hex(
    x: NDArray,
    y: NDArray,
    size: float = 1,
    mode: Literal["default", "flat", "pointy"] = "default",
) -> Tuple[NDArray, NDArray]:
    """Returns hex coordinates from pixel coordinates.

    Args:
        x: Pixel-coordinates in x-direction.
        y: Pixel-coordinates in y-direction.
        size: Size of hexagon.
        mode: Coordinate system convention.

    Returns:
        A tuple containing:
            u: Hex-coordinates in u-direction.
            v: Hex-coordinates in v-direction.

    See Also:
        https://www.redblobgames.com/grids/hexagons/#hex-to-pixel
    """
    if mode == "default":
        return -x / 3 - y / np.sqrt(3), 2 / 3 * x
    elif mode == "flat":
        return (2 / 3 * x) / size, (-1 / 3 * x + np.sqrt(3) / 3 * y) / size
    elif mode == "pointy":
        return (np.sqrt(3) / 3 * x - 1 / 3 * y) / size, (2 / 3 * y) / size
    else:
        raise ValueError(f"{mode} not recognized.")


def pad_to_regular_hex(
    u: NDArray,
    v: NDArray,
    values: NDArray,
    extent: int,
    value: float = np.nan,
) -> Tuple[NDArray, NDArray, NDArray]:
    """Pad hexals with coordinates to a regular hex lattice.

    Args:
        u: U-coordinate of hexal.
        v: V-coordinate of hexal.
        values: Value of hexal with arbitrary shape but last axis
            must match the hexal dimension.
        extent: Extent of regular hex grid to pad to.
        value: The pad value.

    Returns:
        A tuple containing:
            u_padded: Padded u-coordinate.
            v_padded: Padded v-coordinate.
            values_padded: Padded value.

    Note:
        The canonical use case here is to pad a filter, receptive field, or
        postsynaptic current field for visualization.

    Example:
        ```python
        u = np.array([1, 0, -1, 0, 1, 2])
        v = np.array([-2, -1, 0, 0, 0, 0])
        values = np.array([0.05, 0.1, 0.3, 0.5, 0.7, 0.9])
        hexals = pad_to_regular_hex(u, v, values, 6)
        hex_scatter(*hexals, edgecolor='k', cmap=plt.cm.Blues, vmin=0, vmax=1)
        ```
    """
    u_padded, v_padded = flyvis.utils.hex_utils.get_hex_coords(extent)
    slices = tuple()
    if len(values.shape) > 1:
        values_padded = np.ones([*values.shape[:-1], len(u_padded)]) * value
        for _ in range(len(values.shape[:-1])):
            slices += (slice(None),)
    else:
        values_padded = np.ones([len(u_padded)]) * value
    index = flyvis.utils.tensor_utils.where_equal_rows(
        np.stack((u, v), axis=1), np.stack((u_padded, v_padded), axis=1)
    )
    slices += (index,)
    values_padded[slices] = values
    return u_padded, v_padded, values_padded


def max_extent_index(u: NDArray, v: NDArray, max_extent: int) -> NDArray:
    """Returns a mask to constrain u and v axial-hex-coordinates by max_extent.

    Args:
        u: Hex-coordinates in u-direction.
        v: Hex-coordinates in v-direction.
        max_extent: Maximal extent.

    Returns:
        Boolean mask.
    """
    return (
        (-max_extent <= u)
        & (u <= max_extent)
        & (-max_extent <= v)
        & (v <= max_extent)
        & (-max_extent <= u + v)
        & (u + v <= max_extent)
    )


def get_num_hexals(extent: int) -> int:
    """Returns the absolute number of hexals in a hexagonal grid with extent.

    Args:
        extent: Extent of hex-lattice.

    Returns:
        Number of hexals.

    Note:
        Inverse of get_hextent.
    """
    return 1 + 3 * extent * (extent + 1)


def get_hextent(num_hexals: int) -> int:
    """Computes the hex-lattice extent from the number of hexals.

    Args:
        num_hexals: Number of hexals.

    Returns:
        Extent of hex-lattice.

    Note:
        Inverse of get_num_hexals.
    """

    return np.floor(np.sqrt(num_hexals / 3)).astype("int")


def sort_u_then_v(
    u: NDArray, v: NDArray, values: NDArray
) -> Tuple[NDArray, NDArray, NDArray]:
    """Sorts u, v, and values by u and then v.

    Args:
        u: U-coordinate of hexal.
        v: V-coordinate of hexal.
        values: Value of hexal.

    Returns:
        A tuple containing:
            u: Sorted u-coordinate of hexal.
            v: Sorted v-coordinate of hexal.
            values: Sorted value of hexal.
    """
    index = np.lexsort((v, u))
    return u[index], v[index], values[index]


def sort_u_then_v_index(u: NDArray, v: NDArray) -> NDArray:
    """Index to sort u, v by u and then v.

    Args:
        u: U-coordinate of hexal.
        v: V-coordinate of hexal.

    Returns:
        Index to sort u and v.
    """
    return np.lexsort((v, u))


def get_extent(u: NDArray, v: NDArray, astype: type = int) -> int:
    """Returns extent (integer distance to origin) of arbitrary u, v coordinates.

    Args:
        u: U-coordinate of hexal.
        v: V-coordinate of hexal.
        astype: Type to cast to.

    Returns:
        Extent of hex-lattice.

    Note:
        If u and v are arrays, returns the maximum extent.

    See Also:
        https://www.redblobgames.com/grids/hexagons/#distances
    """
    if isinstance(u, Number) and isinstance(v, Number):
        u, v = np.array((u,)), np.array((v,))
    uv = np.stack((u, v), 1)
    extent = (
        abs(0 - uv[:, 0]) + abs(0 + 0 - uv[:, 0] - uv[:, 1]) + abs(0 - uv[:, 1])
    ) / 2
    return np.max(extent).astype(astype)


def crop_to_extent(
    u: NDArray, v: NDArray, color: NDArray, max_extent: int
) -> Tuple[NDArray, NDArray, NDArray]:
    """
    Crop hexagonal grid data to a specified maximum extent.

    Args:
        u: Array of hex coordinates in u direction.
        v: Array of hex coordinates in v direction.
        color: Array of values associated with each (u, v) coordinate.
        max_extent: Maximum extent to crop the hexagonal grid to.

    Returns:
        Tuple of cropped u, v, and color arrays.
    """
    extent_condition = (
        (-max_extent <= u)
        & (u <= max_extent)
        & (-max_extent <= v)
        & (v <= max_extent)
        & (-max_extent <= u + v)
        & (u + v <= max_extent)
    )
    return u[extent_condition], v[extent_condition], color[extent_condition]


# -- Experimental explicit hex-datastructures ----------------------------------


class Hexal:
    """Hexal representation containing u, v, z coordinates and value.

    Attributes:
        u: Coordinate in u principal direction (0 degree axis).
        v: Coordinate in v principal direction (60 degree axis).
        z: Coordinate in z principal direction (-60 degree axis).
        value: 'Hexal' value.
        u_stride: Stride in u-direction.
        v_stride: Stride in v-direction.
    """

    def __init__(
        self, u: int, v: int, value: float = np.nan, u_stride: int = 1, v_stride: int = 1
    ):
        self.u = u
        self.v = v
        self.z = -(u + v)
        self.value = value
        self.u_stride = u_stride
        self.v_stride = v_stride

    def __repr__(self):
        return "Hexal(u={}, v={}, value={}, u_stride={}, v_stride={})".format(
            self.u, self.v, self.value, self.u_stride, self.v_stride
        )

    def __eq__(self, other):
        """Compares coordinates (not values)."""
        if isinstance(other, Hexal):
            return all((self.u == other.u, self.v == other.v))
        elif isinstance(other, Iterable):
            return np.array([self == h for h in other])

    def __add__(self, other):
        """Adds u and v coordinates, while keeping the value of the left hexal."""
        if isinstance(other, Hexal):
            return Hexal(self.u + other.u, self.v + other.v, self.value)
        elif isinstance(other, Iterable):
            return np.array([self + h for h in other])

    def __mul__(self, other):
        """Multiplies values, while preserving coordinates."""
        if isinstance(other, Hexal):
            return Hexal(self.u, self.v, self.value * other.value)
        elif isinstance(other, Iterable):
            return np.array([self * h for h in other])
        else:
            return Hexal(self.u, self.v, self.value * other)

    def eq_val(self, other):
        """Compares the values, not the coordinates."""
        if isinstance(other, Hexal):
            return self.value == other.value
        elif isinstance(other, Iterable):
            return np.array([self.eq_val(h) for h in other])

    # ----- Neighbour identification

    @property
    def east(self):
        return Hexal(self.u + self.u_stride, self.v, 0)

    @property
    def north_east(self):
        return Hexal(self.u, self.v + self.v_stride, 0)

    @property
    def north_west(self):
        return Hexal(self.u - self.u_stride, self.v + self.v_stride, 0)

    @property
    def west(self):
        return Hexal(self.u - self.u_stride, self.v, 0)

    @property
    def south_west(self):
        return Hexal(self.u, self.v - self.v_stride, 0)

    @property
    def south_east(self):
        return Hexal(self.u + self.u_stride, self.v - self.v_stride, 0)

    def neighbours(self):
        """Returns 6 neighbours sorted CCW, starting from east."""
        return (
            self.east,
            self.north_east,
            self.north_west,
            self.west,
            self.south_west,
            self.south_east,
        )

    def is_neighbour(self, other):
        """Evaluates if other is a neighbour."""
        neighbours = self.neighbours()
        if isinstance(other, Hexal):
            return other in neighbours
        elif isinstance(other, Iterable):
            return np.array([self.neighbour(h) for h in other])

    @staticmethod
    def unit_directions():
        """Returns the six unit directions."""
        return HexArray(Hexal(0, 0, 0).neighbours())

    def neighbour(self, angle):
        neighbours = np.array(self.neighbours())
        angles = np.array([h.angle(signed=True) for h in neighbours])
        distance = (angles - angle) % np.pi
        index = np.argsort(distance)
        return HexArray(neighbours[index[:2]])

    def direction(self, angle):
        neighbours = HexArray(self.neighbour(angle))
        angles = np.array([h.angle(signed=True) for h in neighbours])
        distance = (angles - angle) % np.pi
        index = np.argsort(distance)
        return HexArray(self.unit_directions()[index[:2]])

    # ----- Geometric methods

    def interp(self, other, t):
        """Interpolates towards other.

        Args:
            other (Hexal)
            t (float): interpolation step, 0<t<1.

        Returns:
            Hexal
        """

        def hex_round(u, v):
            z = -(u + v)
            ru = round(u)
            rv = round(v)
            rz = round(z)
            u_diff = abs(ru - u)
            v_diff = abs(rv - v)
            z_diff = abs(rz - z)
            if u_diff > v_diff and u_diff > z_diff:
                ru = -rv - rz
            elif v_diff > z_diff:
                rv = -ru - rz
            return ru, rv

        uprime, vprime = (
            self.u + (other.u - self.u) * t,
            self.v + (other.v - self.v) * t,
        )
        uprime, vprime = hex_round(uprime, vprime)
        return Hexal(uprime, vprime, 0)

    def angle(self, other=None, non_negative=False):
        """
        Returns the angle to other or the origin.

        Args:
            other (Hexal)
            non_negative (bool): add 2pi if angle is negative.
                Default: False.

        Returns:
            float: angle in radians.
        """

        def _angle(p1, p2):
            """Counter clockwise angle from p1 to p2.

            Returns:
                float: angle in [0, np.pi]
            """
            dot = p1[0] * p2[0] + p1[1] * p2[1]
            det = p1[0] * p2[1] - p1[1] * p2[0]
            angle = np.arctan2(det, dot)
            return angle

        x, y = self._to_pixel(self.u, self.v)
        theta = np.arctan2(y, x)
        if other is not None:
            xother, yother = self._to_pixel(other.u, other.v)
            theta = _angle([x, y], [xother, yother])
        if non_negative:
            theta += 2 * np.pi if theta < 0 else 0
        return theta

    def distance(self, other=None):
        """Returns the columnar distance between to hexals."""
        if other is not None:
            return int(
                (
                    abs(self.u - other.u)
                    + abs(self.u + self.v - other.u - other.v)
                    + abs(self.v - other.v)
                )
                / 2
            )
        return int((abs(self.u) + abs(self.u + self.v) + abs(self.v)) / 2)

    @staticmethod
    def _to_pixel(u, v, scale=1):
        """Converts to pixel coordinates."""
        return hex_to_pixel(u, v, scale)


class HexArray(np.ndarray):
    """Flat array holding Hexal's as elements.

    Can be constructed with:
        HexArray(hexals: Iterable, values: Optional[np.nan])
        HexArray(u: Iterable, v: Iterable, values: Optional[np.nan])
    """

    def __new__(cls, hexals=None, u=None, v=None, values=0):
        if isinstance(hexals, Iterable):
            u = np.array([h.u for h in hexals])
            v = np.array([h.v for h in hexals])
            values = np.array([h.value for h in hexals])
        if not isinstance(values, Iterable):
            values = np.ones_like(u) * values
        u, v = HexArray.sort(u, v)
        hexals = np.array(
            [Hexal(_u, _v, _val) for _u, _v, _val in zip(u, v, values)],
            dtype=Hexal,
        ).view(cls)
        return hexals

    def __array_finalize__(self, obj):
        if obj is None:
            return

    def __eq__(self, other):
        if isinstance(other, Hexal):
            return other == self
        else:
            return super().__eq__(other)

    def __getitem__(self, key):
        if isinstance(key, HexArray):
            mask = self.where_hexarray(key)
            return self[mask]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if isinstance(key, slice) and key == slice(None):
            self.values = value
        elif isinstance(key, HexArray):
            mask = self.where_hexarray(key)
            super().__setitem__(mask, value)
        else:
            super().__setitem__(key, value)

    def where_hexarray(self, hexarray):
        return matrix_mask_by_sub(
            np.stack((hexarray.u, hexarray.v), axis=0).T,
            np.stack((self.u, self.v), axis=0).T,
        )

    @staticmethod
    def sort(u, v):
        sort_index = np.lexsort((v, u))
        u = u[sort_index]
        v = v[sort_index]
        return u, v

    @staticmethod
    def get_extent(hexals=None, u=None, v=None, center=Hexal(0, 0, 0)):
        """Returns the columnar extent."""
        from numbers import Number

        if isinstance(u, Number) and isinstance(v, Number):
            h = Hexal(u, v, 0)
            return h.distance(center)
        else:
            ha = HexArray(hexals, u, v)
            distance = max([h.distance(center) for h in ha])
            return distance

    @property
    def u(self):
        return np.array([h.u for h in self])

    @property
    def v(self):
        return np.array([h.v for h in self])

    @property
    def values(self):
        return np.array([h.value for h in self])

    @values.setter
    def values(self, values):
        for h, val in zip(self, values):
            h.value = val

    @property
    def extent(self):
        return super().get_extent(self)

    def with_stride(self, u_stride=None, v_stride=None):
        """Returns a sliced instance obeying strides in u- and v-direction."""
        new = []
        for u, v, _ in zip(self.u, self.v, self.values):
            new.append(u % u_stride == 0 and v % v_stride == 0)
        return self[np.array(new)]

    def where(self, value):
        """Returns a mask of where values are equal to the given one.

        Note: value can be np.nan.
        """
        return np.isclose(self.values, value, rtol=0, atol=0, equal_nan=True)

    def fill(self, value):
        """Fills the values with the given one."""
        for h in self:
            h.value = value

    def to_pixel(self, scale=1, mode="default"):
        """Converts to pixel coordinates."""
        return hex_to_pixel(self.u, self.v, scale, mode=mode)

    def plot(self, figsize=[3, 3], fill=True):
        """Plots values in regular hexagonal lattice.

        Meant for debugging.
        """
        u = np.array([h.u for h in self])
        v = np.array([h.v for h in self])
        color = np.array([h.value for h in self])
        return flyvis.plots.hex_scatter(
            u,
            v,
            color,
            fill=fill,
            cmap=cm.get_cmap("binary"),
            edgecolor="black",
            figsize=figsize,
        )


class HexLattice(HexArray):
    """Flat array of Hexals.

    Args:
        extent: Extent of the regular hexagon grid.
        hexals: Existing hexals to initialize with.
        center: Center hexal of the lattice.
        u_stride: Stride in u-direction.
        v_stride: Stride in v-direction.
    """

    def __new__(
        cls,
        extent=15,
        hexals=None,
        center=Hexal(0, 0, 0),
        u_stride=1,
        v_stride=1,
    ):
        if isinstance(hexals, Iterable):
            hexals = HexArray(hexals=hexals)
            u = np.array([h.u for h in hexals])
            v = np.array([h.v for h in hexals])
            extent = extent or super().get_extent(hexals, center=center)
            lattice = HexLattice(
                extent=extent,
                center=center,
                u_stride=u_stride,
                v_stride=v_stride,
            )
            for h in lattice:
                if h in hexals:
                    h.value = hexals[h == hexals][0].value
        else:
            u, v = flyvis.utils.hex_utils.get_hex_coords(extent)
            u += center.u
            v += center.v
            values = [np.nan for _ in range(len(u))]  # np.ones_like(u) * np.nan
            lattice = []
            for _u, _v, _val in zip(u, v, values):
                if _u % u_stride == 0 and _v % v_stride == 0:
                    lattice.append(Hexal(_u, _v, _val, u_stride, v_stride))
            lattice = np.array(lattice, dtype=Hexal).view(cls)
        return lattice

    @property
    def center(self):
        return self[len(self) // 2]

    @property
    def extent(self):
        return super().get_extent(self, center=self.center)

    # ----- Geometry

    def circle(self, radius=None, center=Hexal(0, 0, 0), as_lattice=False):
        """Draws a circle in hex coordinates.

        Args:
            radius: Radius in columns of the circle.
            center: Center of the circle.
            as_lattice: Returns the circle on a constrained regular lattice.
        """
        lattice = HexLattice(extent=max(radius or 0, self.extent), center=center)
        radius = radius or self.extent
        circle = []
        for _, h in enumerate(lattice):
            distance = center.distance(h)
            if distance == radius:
                h.value = 1
                circle.append(h)
        if as_lattice:
            return HexLattice(hexals=circle)
        return HexArray(hexals=circle)

    @staticmethod
    def filled_circle(radius=None, center=Hexal(0, 0, 0), as_lattice=False):
        """Draws a circle in hex coordinates.

        Args:
            radius: Radius in columns of the circle.
            center: Center of the circle.
            as_lattice: Returns the circle on a constrained regular lattice.
        """
        lattice = HexLattice(extent=radius or 0, center=center)
        radius = radius
        circle = []
        for _, h in enumerate(lattice):
            distance = center.distance(h)
            if distance <= radius:
                h.value = 1
                circle.append(h)
        if as_lattice:
            return HexLattice(hexals=circle)
        return HexArray(hexals=circle)

    def hull(self):
        """Returns the hull of the regular lattice."""
        return self.circle(radius=self.extent, center=self.center)

    def _line_span(self, angle):
        """Returns two points spanning a line with given angle wrt. origin.

        Args:
            angle: In [0, np.pi]

        Returns:
            HexArray
        """
        # To offset the line by simple addition of the offset,
        # radius=2 * self.extent spans the line in ways that each valid offset
        # can be added.
        distant_hull = self.ring(radius=2 * self.extent)
        angles = np.array([h.angle(signed=True) for h in distant_hull])
        distance = (angles - angle) % np.pi
        index = np.argsort(distance)
        span = distant_hull[index[0:2]]
        for h in span:
            h.value = 1
        return HexArray(hexals=span)

    def line(self, angle, center=Hexal(0, 0, 1), as_lattice=False):
        """Returns a line on a HexLattice or HexArray.

        Args:
            angle: In [0, np.pi]
            center: Midpoint of the line
            as_lattice: Returns the ring on a constrained regular lattice.

        Returns:
            HexArray or constrained HexLattice
        """
        line_span = self._line_span(angle)
        distance = line_span[0].distance(line_span[1])
        line = []
        for i in range(distance + 1):
            _next = line_span[0].interp(line_span[1], 1 / distance * i)
            line.append(_next)
        for h in line:
            h.value = 1
        if as_lattice:
            return HexLattice(extent=self.extent, hexals=center + line)
        return HexArray(hexals=center + line)

    def _get_neighbour_indices(self, index):
        _neighbours = self[index].neighbours()
        neighbours = ()
        for n in _neighbours:
            valid = self == n
            if valid.any():
                neighbours += (np.where(valid)[0][0],)
        return neighbours

    def valid_neighbours(self):
        neighbours = ()
        for i in range(len(self)):
            neighbours += (self._get_neighbour_indices(i),)
        return neighbours


class LatticeMask:
    """Boolean masks for lattice dimension.

    Args:
        extent: Extent of the hexagonal lattice.
        u_stride: Stride in u-direction.
        v_stride: Stride in v-direction.
    """

    def __init__(self, extent: int = 15, u_stride: int = 1, v_stride: int = 1):
        self._lattice = HexLattice(extent=extent, u_stride=u_stride, v_stride=v_stride)

    @property
    def center(self):
        return self._lattice.center == self._lattice

    @property
    def center_east(self):
        return self._lattice.center.east == self._lattice

    @property
    def center_north_east(self):
        return self._lattice.center.north_east == self._lattice

    @property
    def center_north_west(self):
        return self._lattice.center.north_west == self._lattice

    @property
    def center_west(self):
        return self._lattice.center.west == self._lattice

    @property
    def center_south_west(self):
        return self._lattice.center.south_west == self._lattice

    @property
    def center_south_east(self):
        return self._lattice.center.south_east == self._lattice



import argparse
import os


def replace_in_markdown_files(folder_path, old_path, new_path):
    # Check if the folder_path exists and is a directory
    if not os.path.exists(folder_path) or not os.path.isdir(folder_path):
        print(f"Error: {folder_path} does not exist or is not a directory")
        return

    # Check if the folder_path contains markdown files
    markdown_files = [f for f in os.listdir(folder_path) if f.endswith('.md')]
    if not markdown_files:
        print(f"Error: {folder_path} does not contain any markdown files")
        return

    # Walk through all files in the folder
    for root, _, files in os.walk(folder_path):
        for file in files:
            # Check if file is markdown
            if file.endswith('.md'):
                file_path = os.path.join(root, file)

                # Read file content
                with open(file_path, 'r', encoding='utf-8') as f:
                    content = f.read()

                # Replace the path
                new_content = content.replace(old_path, new_path)

                # Write back only if changes were made
                if new_content != content:
                    with open(file_path, 'w', encoding='utf-8') as f:
                        f.write(new_content)
                    print(f"Updated: {file_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Replace paths in markdown files')
    parser.add_argument(
        'folder_path', help='Path to the folder containing markdown files'
    )
    parser.add_argument(
        '--old_path',
        help='Old path to replace',
        default="/groups/turaga/home/lappalainenj/FlyVis/private/flyvision",
    )
    parser.add_argument(
        '--new_path', help='New path to replace with', default="../flyvis"
    )
    args = parser.parse_args()

    replace_in_markdown_files(args.folder_path, args.old_path, args.new_path)



"""Script to run a jupyter notebook for an ensemble."""

import argparse
import logging
import sys
from importlib import resources
from typing import List

from flyvis.utils.compute_cloud_utils import launch_single

logging.basicConfig(
    format="[%(asctime)s] [%(filename)s:%(lineno)d] %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

NOTEBOOK_SCRIPT_DIR = resources.files("flyvis_cli") / "analysis/notebook.py"


def run_ensemble_notebook(args: argparse.Namespace, kwargs: List[str]) -> None:
    """
    Launch ensemble notebook job.

    Args:
        args: Command-line arguments.
        kwargs: Additional keyword arguments as a list of strings.
    """
    launch_single(
        args.ensemble_id,
        args.task_name,
        args.nP,
        args.gpu,
        args.q,
        str(NOTEBOOK_SCRIPT_DIR),
        args.dry,
        ["--notebook_path", args.notebook_path] + ["per_ensemble:bool=true"] + kwargs,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Run ensemble notebook on the compute cloud.",
        usage=(
            "\nflyvis notebook-per-ensemble [-h] [...] --ensemble_id ENSEMBLE_ID "
            "--task_name TASK_NAME --notebook_path PATH\n"
            "       or\n"
            "%(prog)s [-h] [...] --ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "--notebook_path PATH\n"
        ),
    )
    parser.add_argument("--nP", type=int, default=4, help="Number of processors.")
    parser.add_argument("--gpu", type=str, default="num=1", help="Number of GPUs.")
    parser.add_argument("--q", type=str, default="gpu_l4", help="Queue.")
    parser.add_argument(
        "--ensemble_id",
        type=int,
        required=True,
        help="Id of the ensemble, e.g. 0045.",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        required=True,
        help="Name given to the task, e.g., flow.",
    )
    DEFAULT_NOTEBOOK_PATH = str(
        resources.files("flyvis.analysis").joinpath("__main__.ipynb")
    )
    parser.add_argument(
        "--notebook_path",
        type=str,
        default=DEFAULT_NOTEBOOK_PATH,
        help=f"Path of the notebook to execute. Default: {DEFAULT_NOTEBOOK_PATH}",
    )
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Perform a dry run without actually launching jobs.",
    )

    args, kwargs = parser.parse_known_intermixed_args()
    run_ensemble_notebook(args, sys.argv[1:])



"""Script to validate a single network across all its checkpoints."""

import argparse

from flyvis import NetworkView
from flyvis.analysis.validation import validate_all_checkpoints
from flyvis.utils.config_utils import HybridArgumentParser


def main():
    parser = HybridArgumentParser(
        hybrid_args={
            'ensemble_and_network_id': {
                'required': True,
                'help': 'ID of the ensemble and network to use, e.g. 0045/000',
            },
            'task_name': {
                'required': True,
                'help': (
                    'Name of the task. Resulting network name will be '
                    'task_name/ensemble_and_network_id.'
                ),
            },
        },
        description=(
            "Validate a single network across all its checkpoints. "
            "Computes and stores validation metrics in the network's validation "
            "directory."
        ),
        usage=(
            "\nflyvis val-single [-h] task_name=TASK ensemble_and_network_id=XXXX/YYY\n"
            "       or\n"
            "%(prog)s [-h] task_name=TASK ensemble_and_network_id=XXXX/YYY\n"
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
--------
1. Validate a specific network:
    flyvis val-single task_name=flow ensemble_and_network_id=0045/000

2. Validate a network from a different task:
    flyvis val-single task_name=depth ensemble_and_network_id=0023/012
""",
    )

    args = parser.parse_with_hybrid_args()

    network_name = f"{args.task_name}/{args.ensemble_and_network_id}"
    network_view = NetworkView(network_name)

    # TODO: add logic to pass different validation methods
    validate_all_checkpoints(network_view)


if __name__ == "__main__":
    main()



"""Interface to control the cell-specific stimulus buffer for the network."""

from typing import Any, Callable, Dict, Optional, Protocol, Type, Union, runtime_checkable

import numpy as np
import torch
from numpy.typing import NDArray
from torch import Tensor

from flyvis.connectome import ConnectomeFromAvgFilters

__all__ = ["Stimulus"]


@runtime_checkable
class StimulusProtocol(Protocol):
    """Protocol for the Stimulus class."""

    def __call__(self) -> Tensor: ...
    def add_input(self, x: Tensor, **kwargs) -> None: ...
    def add_pre_stim(self, x: Tensor, **kwargs) -> None: ...
    def zero(self, **kwargs) -> None: ...
    def nonzero(self) -> bool: ...


AVAILABLE_STIMULI: Dict[str, Type[StimulusProtocol]] = {}


def register_stimulus(
    cls: Optional[Type[StimulusProtocol]] = None,
) -> Union[
    Callable[[Type[StimulusProtocol]], Type[StimulusProtocol]], Type[StimulusProtocol]
]:
    """Register a stimulus class.

    Args:
        cls: The stimulus class to register (optional when used as a decorator).

    Returns:
        Registered class or decorator function.

    Example:
        As a standalone function:
        ```python
        register_stimulus(CustomStimulus)
        ```

        As a decorator:
        ```python
        @register_stimulus
        class CustomStimulus(StimulusProtocol): ...
        ```
    """

    def decorator(cls: Type[StimulusProtocol]) -> Type[StimulusProtocol]:
        AVAILABLE_STIMULI[cls.__name__] = cls
        return cls

    if cls is None:
        return decorator
    else:
        return decorator(cls)


@register_stimulus
class Stimulus:
    """Interface to control the cell-specific stimulus buffer for the network.

    Creates a buffer and maps standard video input to the photoreceptors
    but can map input to any other cell as well, e.g. to do perturbation
    experiments.

    Args:
        connectome: Connectome directory to retrieve indexes for the stimulus
            buffer at the respective cell positions.
        n_samples: Number of samples to initialize the buffer with.
        n_frames: Number of frames to initialize the buffer with.
        init_buffer: If False, do not initialize the stimulus buffer.

    Attributes:
        layer_index (Dict[str, NDArray]): Dictionary of cell type to index array.
        central_cells_index (Dict[str, int]): Dictionary of cell type to
            central cell index.
        input_index (NDArray): Index array of photoreceptors.
        n_frames (int): Number of frames in the stimulus buffer.
        n_samples (int): Number of samples in the stimulus buffer.
        n_nodes (int): Number of nodes in the stimulus buffer.
        n_input_elements (int): Number of input elements.
        buffer (Tensor): Stimulus buffer of shape (n_samples, n_frames, n_cells).

    Returns:
        Tensor: Stimulus of shape (n_samples, n_frames, n_cells)

    Example:
        ```python
        stim = Stimulus(network.connectome, *x.shape[:2])
        stim.add_input(x)
        response = network(stim(), dt)
        ```
    """

    layer_index: Dict[str, NDArray]
    central_cells_index: Dict[str, int]
    input_index: NDArray
    n_frames: int
    n_samples: int
    n_nodes: int
    n_input_elements: int
    buffer: Tensor

    def __init__(
        self,
        connectome: ConnectomeFromAvgFilters,
        n_samples: int = 1,
        n_frames: int = 1,
        init_buffer: bool = True,
    ):
        self.layer_index = {
            cell_type: index[:]
            for cell_type, index in connectome.nodes.layer_index.items()
        }
        self.central_cells_index = dict(
            zip(
                connectome.unique_cell_types[:].astype(str),
                connectome.central_cells_index[:],
            )
        )
        self.input_index = np.array([
            self.layer_index[cell_type.decode()]
            for cell_type in connectome.input_cell_types[:]
        ])
        self.n_input_elements = self.input_index.shape[1]
        self.n_samples, self.n_frames, self.n_nodes = (
            n_samples,
            n_frames,
            len(connectome.nodes.type),
        )
        self.connectome = connectome
        if init_buffer:
            self.zero()

    def zero(
        self,
        n_samples: Optional[int] = None,
        n_frames: Optional[int] = None,
    ) -> None:
        """Reset the stimulus buffer to zero.

        Args:
            n_samples: Number of samples. If provided, the buffer will be resized.
            n_frames: Number of frames. If provided, the buffer will be resized.
        """
        self.n_samples = n_samples or self.n_samples
        self.n_frames = n_frames or self.n_frames
        if hasattr(self, "buffer") and self.buffer.shape[:2] == (
            self.n_samples,
            self.n_frames,
        ):
            self.buffer.zero_()
            return
        self.buffer = torch.zeros((self.n_samples, self.n_frames, self.n_nodes))
        self._nonzero = False

    @property
    def nonzero(self) -> bool:
        """Check if elements have been added to the stimulus buffer.

        Returns:
            bool: True if elements have been added, even if those elements were all zero.
        """
        return self._nonzero

    def add_input(
        self,
        x: torch.Tensor,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        n_frames_buffer: Optional[int] = None,
        cumulate: bool = False,
    ) -> None:
        """Add input to the input/photoreceptor cells.

        Args:
            x: Input video of shape (n_samples, n_frames, 1, n_input_elements).
            start: Temporal start index of the stimulus.
            stop: Temporal stop index of the stimulus.
            n_frames_buffer: Number of frames to resize the buffer to.
            cumulate: If True, add input to the existing buffer.

        Raises:
            ValueError: If input shape is incorrect.
            RuntimeError: If input shape doesn't match buffer shape.
        """
        shape = x.shape
        if len(shape) != 4:
            raise ValueError(
                f"input has shape {x.shape} but must have "
                "(n_samples, n_frames, 1, n_input_elements)"
            )
        n_samples, n_frames_input = shape[:2]

        if not hasattr(self, "buffer") or not cumulate and self.nonzero:
            self.zero(n_samples, n_frames_buffer or n_frames_input)

        try:
            self.buffer[:, slice(start, stop), self.input_index] += x.to(
                self.buffer.device
            )
        except RuntimeError as e:
            raise RuntimeError(
                f"input has shape {x.shape} but buffer has shape {self.buffer.shape}"
            ) from e
        self._nonzero = True

    def add_pre_stim(
        self,
        x: torch.Tensor,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        n_frames_buffer: Optional[int] = None,
    ) -> None:
        """Add a constant or sequence of constants to the input/photoreceptor cells.

        Args:
            x: Grey value(s). If Tensor, must have length `n_frames` or `stop - start`.
            start: Start index in time.
            stop: Stop index in time.
            n_frames_buffer: Number of frames to resize the buffer to.

        Raises:
            RuntimeError: If input shape doesn't match buffer shape.
        """
        if not hasattr(self, "buffer") or self.nonzero:
            self.zero(None, n_frames_buffer)

        try:
            if isinstance(x, torch.Tensor) and x.ndim == 1:
                self.buffer[:, slice(start, stop), self.input_index] += x.view(
                    1, len(x), 1, 1
                )
            else:
                self.buffer[:, slice(start, stop), self.input_index] += x
        except RuntimeError as e:
            raise RuntimeError(
                f"input has shape {x.shape} but buffer has shape {self.buffer.shape}"
            ) from e
        self._nonzero = True

    def __call__(self) -> torch.Tensor:
        """Return the stimulus tensor.

        Returns:
            torch.Tensor: The stimulus buffer.
        """
        return self.buffer


def is_stimulus_protocol(obj: Any) -> bool:
    return isinstance(obj, StimulusProtocol)


def init_stimulus(connectome: ConnectomeFromAvgFilters, **kwargs) -> StimulusProtocol:
    if "type" not in kwargs:
        return None
    stimulus_class = AVAILABLE_STIMULI[kwargs.pop("type")]
    return stimulus_class(connectome, **kwargs)



# pylint: disable=dangerous-default-value
"""
Deep mechanistic network module.
"""

from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from functools import wraps
from os import PathLike
from pprint import pformat
from typing import Any, Callable, Dict, List, Optional, Union

import torch.nn as nn
import xarray as xr
from cachetools import FIFOCache
from datamate import Directory, set_root_context
from joblib import Memory

import flyvis
from flyvis.analysis import (
    optimal_stimuli,
    stimulus_responses,
    stimulus_responses_currents,
)
from flyvis.connectome import (
    ConnectomeView,
    get_avgfilt_connectome,
)
from flyvis.task.tasks import init_decoder
from flyvis.utils.cache_utils import context_aware_cache, make_hashable
from flyvis.utils.chkpt_utils import (
    best_checkpoint_default_fn,
    recover_decoder,
    recover_network,
    resolve_checkpoints,
)

from .directories import NetworkDir
from .network import Network

logger = logging.getLogger(__name__)

__all__ = ["NetworkView", "CheckpointedNetwork"]


class NetworkView:
    """IO interface for network.

    Args:
        network_dir: Directory of the network.
        network_class: Network class. Defaults to Network.
        root_dir: Root directory. Defaults to flyvis.results_dir.
        connectome_getter: Function to get the connectome.
            Defaults to flyvision_connectome.
        checkpoint_mapper: Function to map checkpoints. Defaults to resolve_checkpoints.
        best_checkpoint_fn: Function to get the best checkpoint. Defaults to
            best_checkpoint_default_fn.
        best_checkpoint_fn_kwargs: Keyword arguments for best_checkpoint_fn. Defaults to
            {"validation_subdir": "validation", "loss_file_name": "loss"}.
        recover_fn: Function to recover the network. Defaults to recover_network.

    Attributes:
        network_class (nn.Module): Network class.
        dir (Directory): Network directory.
        name (str): Network name.
        root_dir (PathLike): Root directory.
        connectome_getter (Callable): Function to get the connectome.
        checkpoint_mapper (Callable): Function to map checkpoints.
        connectome_view (ConnectomeView): Connectome view.
        connectome (Directory): Connectome directory.
        checkpoints: Mapped checkpoints.
        memory (Memory): Joblib memory cache.
        best_checkpoint_fn (Callable): Function to get the best checkpoint.
        best_checkpoint_fn_kwargs (dict): Keyword arguments for best_checkpoint_fn.
        recover_fn (Callable): Function to recover the network.
        _network (CheckpointedNetwork): Checkpointed network instance.
        decoder: Decoder instance.
        _initialized (dict): Initialization status for network and decoder.
        cache (FIFOCache): Cache for storing results.
    """

    def __init__(
        self,
        network_dir: Union[str, PathLike, NetworkDir],
        network_class: nn.Module = Network,
        root_dir: PathLike = flyvis.results_dir,
        connectome_getter: Callable = get_avgfilt_connectome,
        checkpoint_mapper: Callable = resolve_checkpoints,
        best_checkpoint_fn: Callable = best_checkpoint_default_fn,
        best_checkpoint_fn_kwargs: dict = {
            "validation_subdir": "validation",
            "loss_file_name": "epe",
        },
        recover_fn: Callable = recover_network,
    ):
        self.network_class = network_class
        self.dir, self.name = self._resolve_dir(network_dir, root_dir)
        self.root_dir = root_dir
        self.connectome_getter = connectome_getter
        self.checkpoint_mapper = checkpoint_mapper
        self.connectome_view: ConnectomeView = connectome_getter(
            self.dir.config.network.connectome
        )
        self.connectome = self.connectome_view.dir
        self.checkpoints = checkpoint_mapper(self.dir)
        self.memory = Memory(
            location=self.dir.path / "__cache__",
            backend="xarray_dataset_h5",
            verbose=0,
            # verbose=11,
        )
        self.best_checkpoint_fn = best_checkpoint_fn
        self.best_checkpoint_fn_kwargs = best_checkpoint_fn_kwargs
        self.recover_fn = recover_fn
        self._network_instance = None
        self.decoder = None
        self._initialized = {"network": None, "decoder": None}
        self.cache = FIFOCache(maxsize=3)
        logging.info("Initialized network view at %s", str(self.dir.path))

    @property
    def _network(self) -> CheckpointedNetwork:
        """Lazy init of CheckpointedNetwork because get_checkpoint can be slow."""
        if self._network_instance is None:
            self._network_instance = CheckpointedNetwork(
                self.network_class,
                self.dir.config.network.to_dict(),
                self.name,
                self.get_checkpoint("best"),
                self.recover_fn,
                network=None,
            )
        return self._network_instance

    @_network.setter
    def _network(self, value):
        """Setter for _network property."""
        self._network_instance = value

    def _clear_cache(self):
        """Clear the FIFO cache."""
        self.cache = self.cache.__class__(maxsize=self.cache.maxsize)

    def _clear_memory(self):
        """Clear the joblib memory cache."""
        self.memory.clear()

    # --- ConnectomeView API for static code analysis
    # pylint: disable=missing-function-docstring
    @wraps(ConnectomeView.connectivity_matrix)
    def connectivity_matrix(self, *args, **kwargs):
        return self.connectome_view.connectivity_matrix(*args, **kwargs)

    connectivity_matrix.__doc__ = ConnectomeView.connectivity_matrix.__doc__

    @wraps(ConnectomeView.network_layout)
    def network_layout(self, *args, **kwargs):
        return self.connectome_view.network_layout(*args, **kwargs)

    network_layout.__doc__ = ConnectomeView.network_layout.__doc__

    @wraps(ConnectomeView.hex_layout)
    def hex_layout(self, *args, **kwargs):
        return self.connectome_view.hex_layout(*args, **kwargs)

    hex_layout.__doc__ = ConnectomeView.hex_layout.__doc__

    @wraps(ConnectomeView.hex_layout_all)
    def hex_layout_all(self, *args, **kwargs):
        return self.connectome_view.hex_layout_all(*args, **kwargs)

    hex_layout_all.__doc__ = ConnectomeView.hex_layout_all.__doc__

    @wraps(ConnectomeView.get_uv)
    def get_uv(self, *args, **kwargs):
        return self.connectome_view.get_uv(*args, **kwargs)

    get_uv.__doc__ = ConnectomeView.get_uv.__doc__

    @wraps(ConnectomeView.sources_list)
    def sources_list(self, *args, **kwargs):
        return self.connectome_view.sources_list(*args, **kwargs)

    sources_list.__doc__ = ConnectomeView.sources_list.__doc__

    @wraps(ConnectomeView.targets_list)
    def targets_list(self, *args, **kwargs):
        return self.connectome_view.targets_list(*args, **kwargs)

    targets_list.__doc__ = ConnectomeView.targets_list.__doc__

    @wraps(ConnectomeView.receptive_field)
    def receptive_field(self, *args, **kwargs):
        return self.connectome_view.receptive_field(*args, **kwargs)

    receptive_field.__doc__ = ConnectomeView.receptive_field.__doc__

    @wraps(ConnectomeView.receptive_fields_grid)
    def receptive_fields_grid(self, *args, **kwargs):
        return self.connectome_view.receptive_fields_grid(*args, **kwargs)

    receptive_fields_grid.__doc__ = ConnectomeView.receptive_fields_grid.__doc__

    @wraps(ConnectomeView.projective_field)
    def projective_field(self, *args, **kwargs):
        return self.connectome_view.projective_field(*args, **kwargs)

    projective_field.__doc__ = ConnectomeView.projective_field.__doc__

    @wraps(ConnectomeView.projective_fields_grid)
    def projective_fields_grid(self, *args, **kwargs):
        return self.connectome_view.projective_fields_grid(*args, **kwargs)

    projective_fields_grid.__doc__ = ConnectomeView.projective_fields_grid.__doc__

    @wraps(ConnectomeView.receptive_fields_df)
    def receptive_fields_df(self, *args, **kwargs):
        return self.connectome_view.receptive_fields_df(*args, **kwargs)

    receptive_fields_df.__doc__ = ConnectomeView.receptive_fields_df.__doc__

    @wraps(ConnectomeView.receptive_fields_sum)
    def receptive_fields_sum(self, *args, **kwargs):
        return self.connectome_view.receptive_fields_sum(*args, **kwargs)

    receptive_fields_sum.__doc__ = ConnectomeView.receptive_fields_sum.__doc__

    @wraps(ConnectomeView.projective_fields_df)
    def projective_fields_df(self, *args, **kwargs):
        return self.connectome_view.projective_fields_df(*args, **kwargs)

    projective_fields_df.__doc__ = ConnectomeView.projective_fields_df.__doc__

    @wraps(ConnectomeView.projective_fields_sum)
    def projective_fields_sum(self, *args, **kwargs):
        return self.connectome_view.projective_fields_sum(*args, **kwargs)

    projective_fields_sum.__doc__ = ConnectomeView.projective_fields_sum.__doc__

    # --- own API

    def get_checkpoint(self, checkpoint="best"):
        """Return the best checkpoint index.

        Args:
            checkpoint: Checkpoint identifier. Defaults to "best".

        Returns:
            str: Path to the checkpoint.
        """
        try:
            if checkpoint == "best":
                return self.best_checkpoint_fn(
                    self.dir.path,
                    **self.best_checkpoint_fn_kwargs,
                )
            return self.checkpoints.paths[checkpoint]
        except FileNotFoundError:
            logger.warning("Checkpoint %s not found at %s", checkpoint, self.dir.path)
            return None

    def network(
        self, checkpoint="best", network: Optional[Any] = None, lazy=False
    ) -> CheckpointedNetwork:
        """Lazy loading of network instance.

        Args:
            checkpoint: Checkpoint identifier. Defaults to "best".
            network: Existing network instance to use. Defaults to None.
            lazy: If True, don't recover the network immediately. Defaults to False.

        Returns:
            CheckpointedNetwork: Checkpointed network instance.
        """
        self._network = CheckpointedNetwork(
            self.network_class,
            self.dir.config.network.to_dict(),
            self.name,
            self.get_checkpoint(checkpoint),
            self.recover_fn,
            network=network or self._network.network,
        )
        if self._network.network is not None and not lazy:
            self._network.recover()
        return self._network

    def init_network(self, checkpoint="best", network: Optional[Any] = None) -> Network:
        """Initialize the network.

        Args:
            checkpoint: Checkpoint identifier. Defaults to "best".
            network: Existing network instance to use. Defaults to None.

        Returns:
            Network: Initialized network instance.
        """
        checkpointed_network = self.network(checkpoint=checkpoint, network=network)

        if checkpointed_network.network is not None:
            return checkpointed_network.network
        checkpointed_network.init()
        return checkpointed_network.recover()

    def init_decoder(self, checkpoint="best", decoder=None):
        """Initialize the decoder.

        Args:
            checkpoint: Checkpoint identifier. Defaults to "best".
            decoder: Existing decoder instance to use. Defaults to None.

        Returns:
            Decoder: Initialized decoder instance.
        """
        checkpointed_network = self.network(checkpoint=checkpoint, lazy=True)
        if (
            self._initialized["decoder"] == checkpointed_network.checkpoint
            and decoder is None
        ):
            return self.decoder
        self.decoder = decoder or init_decoder(
            self.dir.config.task.decoder, self.connectome
        )
        recover_decoder(self.decoder, checkpointed_network.checkpoint)
        self._initialized["decoder"] = checkpointed_network.checkpoint
        return self.decoder

    def _resolve_dir(self, network_dir, root_dir):
        """Resolve the network directory.

        Args:
            network_dir: Network directory path or Directory instance.
            root_dir: Root directory path.

        Returns:
            tuple: (Directory, str) - Network directory and name.

        Raises:
            ValueError: If the directory is not a NetworkDir.
        """
        if isinstance(network_dir, (PathLike, str)):
            with set_root_context(root_dir):
                network_dir = Directory(network_dir)
        if not network_dir.config.type == "NetworkDir":
            raise ValueError(f"NetworkDir not found at {network_dir.path}.")
        name = os.path.sep.join(network_dir.path.parts[-3:])
        return network_dir, name

    # --- stimulus responses

    @wraps(stimulus_responses.flash_responses)
    @context_aware_cache
    def flash_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate flash responses."""
        return stimulus_responses.flash_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.moving_edge_responses)
    @context_aware_cache
    def moving_edge_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate moving edge responses."""
        return stimulus_responses.moving_edge_responses(self, *args, **kwargs)

    @wraps(stimulus_responses_currents.moving_edge_currents)
    @context_aware_cache
    def moving_edge_currents(
        self, *args, **kwargs
    ) -> List[stimulus_responses_currents.ExperimentData]:
        """Generate moving edge currents."""
        return stimulus_responses_currents.moving_edge_currents(self, *args, **kwargs)

    @wraps(stimulus_responses.moving_bar_responses)
    @context_aware_cache
    def moving_bar_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate moving bar responses."""
        return stimulus_responses.moving_bar_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.naturalistic_stimuli_responses)
    @context_aware_cache
    def naturalistic_stimuli_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate naturalistic stimuli responses."""
        return stimulus_responses.naturalistic_stimuli_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.central_impulses_responses)
    @context_aware_cache
    def central_impulses_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate central ommatidium impulses responses."""
        return stimulus_responses.central_impulses_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.spatial_impulses_responses)
    @context_aware_cache
    def spatial_impulses_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate spatial ommatidium impulses responses."""
        return stimulus_responses.spatial_impulses_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.optimal_stimulus_responses)
    @context_aware_cache
    def optimal_stimulus_responses(
        self, cell_type, *args, **kwargs
    ) -> optimal_stimuli.RegularizedOptimalStimulus:
        """Generate optimal stimuli responses."""
        return stimulus_responses.optimal_stimulus_responses(
            self, cell_type, *args, **kwargs
        )


@dataclass
class CheckpointedNetwork:
    """A network representation with checkpoint that can be pickled.

    Attributes:
        network_class: Network class (e.g., flyvis.Network).
        config: Configuration for the network.
        name: Name of the network.
        checkpoint: Checkpoint path.
        recover_fn: Function to recover the network.
        network: Network instance to avoid reinitialization.
    """

    network_class: Any
    config: Dict
    name: str
    checkpoint: PathLike
    recover_fn: Any = recover_network
    network: Optional[Network] = None

    def init(self, eval: bool = True) -> Network:
        """Initialize the network.

        Args:
            eval: Whether to set the network in evaluation mode.

        Returns:
            The initialized network.
        """
        if self.network is None:
            self.network = self.network_class(**self.config)
        if eval:
            self.network.eval()
        return self.network

    def recover(self, checkpoint: Optional[PathLike] = None) -> Network:
        """Recover the network from the checkpoint.

        Args:
            checkpoint: Path to the checkpoint. If None, uses the default checkpoint.

        Returns:
            The recovered network.

        Note:
            Initializes the network if it hasn't been initialized yet.
        """
        if self.network is None:
            self.init()
        return self.recover_fn(self.network, checkpoint or self.checkpoint)

    def __repr__(self):
        return (
            f"CheckpointedNetwork(\n"
            f"    network_class={self.network_class.__name__},\n"
            f"    name='{self.name}',\n"
            f"    config={pformat(self.config, indent=4)},\n"
            f"    checkpoint='{os.path.basename(str(self.checkpoint))}'\n"
            f"    recover_fn={self.recover_fn.__name__}\n"
            f")"
        )

    def _hash_key(self):
        return (
            self.network_class,
            self.name,
            make_hashable(self.config),
            self.checkpoint,
            self.recover_fn.__name__,
        )

    def __hash__(self):
        return hash(self._hash_key())

    # Equality check based on hashable elements.
    def __eq__(self, other):
        if not isinstance(other, CheckpointedNetwork):
            return False
        return (
            self.network_class == other.network_class
            and self.name == other.name
            and make_hashable(self.config) == make_hashable(other.config)
            and self.checkpoint == other.checkpoint
            and self.recover_fn.__name__ == other.recover_fn.__name__
        )

    # Custom reduce method to make the object compatible with joblib's pickling.
    # This ensures the 'network' attribute is never pickled.
    # Return a tuple containing:
    # 1. A callable that will recreate the object (here, the class itself)
    # 2. The arguments required to recreate the object (excluding the network)
    # 3. The state, excluding the 'network' attribute
    def __reduce__(self):
        state = self.__dict__.copy()
        state["network"] = None  # Exclude the complex network from being pickled

        return (
            self.__class__,
            (
                self.network_class,
                self.config,
                self.checkpoint,
                self.recover_fn.__name__,
                None,
            ),
            state,
        )

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.network = None



import logging
from typing import Dict

import torch
from datamate import Namespace
from toolz import valmap
from torch.utils.data import DataLoader, sampler

from flyvis.connectome import ConnectomeFromAvgFilters
from flyvis.datasets.datasets import MultiTaskDataset
from flyvis.utils.class_utils import forward_subclass
from flyvis.utils.dataset_utils import IndexSampler

from . import objectives
from .decoder import ActivityDecoder


class Task:
    """Defines a task for a multi-task dataset from configurations.

    Args:
        dataset: Configuration for the dataset.
        decoder: Configuration for the decoder.
        loss: Configuration for the loss functions.
        batch_size: Size of each batch. Defaults to 4.
        n_iters: Number of iterations. Defaults to 250,000.
        n_folds: Number of folds for cross-validation. Defaults to 4.
        fold: Current fold number. Defaults to 1.
        seed: Random seed for reproducibility. Defaults to 0.
        original_split: Whether to use the original data split. Defaults to False.

    Attributes:
        batch_size: Size of each batch.
        n_iters: Number of iterations.
        n_folds: Number of folds for cross-validation.
        fold: Current fold number.
        seed: Random seed for reproducibility.
        decoder: Configuration for the decoder.
        dataset (MultiTaskDataset): The initialized multi-task dataset.
        losses (Namespace): Loss functions for each task.
        train_seq_index (List[int]): Indices of training sequences.
        val_seq_index (List[int]): Indices of validation sequences.
        train_data (DataLoader): DataLoader for training data.
        train_batch (DataLoader): DataLoader for a single training batch.
        val_data (DataLoader): DataLoader for validation data.
        val_batch (DataLoader): DataLoader for a single validation batch.
        overfit_data (DataLoader): DataLoader for overfitting on a single sample.
    """

    def __init__(
        self,
        dataset: Namespace,
        decoder: Namespace,
        loss: Namespace,
        task_weights: Dict[str, float] = None,
        batch_size: int = 4,
        n_iters: int = 250_000,
        n_folds: int = 4,
        fold: int = 1,
        seed: int = 0,
        original_split: bool = False,
    ):
        self.batch_size = batch_size
        self.n_iters = n_iters
        self.n_folds = n_folds
        self.fold = fold
        self.seed = seed
        self.decoder = decoder

        # Initialize dataset.
        self.dataset = forward_subclass(MultiTaskDataset, dataset)
        self.task_weights, self.task_weights_sum = self.init_task_weights(task_weights)

        self.losses = Namespace({
            task: getattr(objectives, config) for task, config in loss.items()
        })

        if original_split:
            self.train_seq_index, self.val_seq_index = (
                self.dataset.original_train_and_validation_indices()
            )
        else:
            self.train_seq_index, self.val_seq_index = self.dataset.get_random_data_split(
                fold, n_folds, seed
            )

        # Initialize dataloaders.
        self.train_data = DataLoader(
            self.dataset,
            batch_size=batch_size,
            sampler=sampler.SubsetRandomSampler(self.train_seq_index),
            drop_last=True,
        )
        self.train_batch = DataLoader(
            self.dataset,
            batch_size=batch_size,
            sampler=IndexSampler(self.train_seq_index[:batch_size]),
            drop_last=False,
        )
        logging.info(
            "Initialized dataloader with training sequence indices \n%s",
            self.train_seq_index,
        )

        self.val_data = DataLoader(
            self.dataset,
            batch_size=1,
            sampler=IndexSampler(self.val_seq_index),
        )
        self.val_batch = DataLoader(
            self.dataset,
            batch_size=batch_size,
            sampler=IndexSampler(self.val_seq_index[:batch_size]),
        )
        logging.info(
            "Initialized dataloader with validation sequence indices \n%s",
            self.val_seq_index,
        )

        # Initialize overfitting loader.
        self.overfit_data = DataLoader(self.dataset, sampler=IndexSampler([0]))

    def init_decoder(
        self, connectome: ConnectomeFromAvgFilters
    ) -> Dict[str, ActivityDecoder]:
        """Initialize the decoder.

        Args:
            connectome: The connectome directory.

        Returns:
            A dictionary of initialized decoders.
        """
        return init_decoder(self.decoder, connectome)

    def loss(
        self, input: torch.Tensor, target: torch.Tensor, task: str, **kwargs
    ) -> torch.Tensor:
        """Returns the task loss multiplied with the task weight.

        Args:
            input: Input tensor.
            target: Target tensor.
            task: Task name.
            **kwargs: Additional keyword arguments for the loss function.

        Returns:
            Weighted task loss.
        """
        return (
            self.task_weights[task]
            * self.losses[task](input, target, **kwargs)
            / self.task_weights_sum
        )

    def init_task_weights(self, task_weights: Dict[str, float]) -> Dict[str, float]:
        """Returns the task weights.

        Returns:
            A dictionary of task weights.
        """
        task_weights = (
            task_weights
            if task_weights is not None
            else {task: 1 for task in self.dataset.tasks}
        )

        return task_weights, sum(task_weights.values())


def init_decoder(
    config: Dict, connectome: ConnectomeFromAvgFilters
) -> Dict[str, ActivityDecoder]:
    """Initialize decoders.

    Args:
        config: Configuration for the decoders.
        connectome: The connectome directory.

    Returns:
        A dictionary of decoders.

    Example:
        ```python
        decoder = Namespace(
            flow=Namespace(
                type="DecoderGAVP",
                shape=[8, 2],
                kernel_size=5,
                const_weight=0.001,
                p_dropout=0.5,
            ),
            depth=Namespace(
                type="DecoderGAVP",
                shape=[8, 1],
                kernel_size=5,
                const_weight=0.001,
                p_dropout=0.5,
            ),
            lum=Namespace(
                type="DecoderGAVP",
                shape=[8, 3],
                n_out_features=2,
                kernel_size=5,
                const_weight=0.001,
                p_dropout=0.5,
            ),
            shared=False,
        )
        ```
    """
    config = config.deepcopy()

    def init(conf):
        return forward_subclass(ActivityDecoder, {**conf, "connectome": connectome})

    decoder = valmap(init, config)

    return decoder



"""Utility functions for the flyvis package.

This module organizes various utility submodules and sets up custom
accessors and backends.
"""

from importlib import import_module


def __getattr__(name):
    # This lazy loading mechanism is implemented to improve import performance
    # by deferring the import of submodules until they are actually needed.
    # It reduces the initial import time of the utils package, especially
    # beneficial for large codebases or when only specific utilities are required.
    if name in (
        'activity_utils',
        'cache_utils',
        'chkpt_utils',
        'class_utils',
        'color_utils',
        'compute_cloud_utils',
        'config_utils',
        'dataset_utils',
        'df_utils',
        'groundtruth_utils',
        'hex_utils',
        'logging_utils',
        'log_utils',
        'nn_utils',
        'nodes_edges_utils',
        'tensor_utils',
        'type_utils',
        'xarray_joblib_backend',
        'xarray_utils',
    ):
        return import_module(f'flyvis.utils.{name}')
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def setup_xarray_accessors():
    import xarray as xr
    from . import xarray_utils

    if not hasattr(xr.DataArray, "custom"):
        xr.register_dataarray_accessor("custom")(xarray_utils.CustomAccessor)
    if not hasattr(xr.Dataset, "custom"):
        xr.register_dataset_accessor("custom")(xarray_utils.CustomAccessor)


def setup_joblib_backend():
    from joblib import register_store_backend
    from . import xarray_joblib_backend

    register_store_backend(
        'xarray_dataset_h5', xarray_joblib_backend.H5XArrayDatasetStoreBackend
    )


# Setup functions are called here, but they can be moved to be called only when needed
setup_xarray_accessors()
setup_joblib_backend()



import json
import logging
import subprocess
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
from typing import Any


@lru_cache(100)
def warn_once(logger: logging.Logger, msg: str) -> None:
    """
    Log a warning message only once for a given logger and message combination.

    Args:
        logger: The logger object to use for logging.
        msg: The warning message to log.

    Note:
        This function uses an LRU cache to ensure each unique combination of
        logger and message is only logged once.
    """
    logger.warning(msg)


def save_conda_environment(path: Path) -> None:
    """
    Save the current Conda environment to a JSON file.

    Args:
        path: The path where the JSON file will be saved.

    Note:
        The function appends '.json' to the provided path.
    """
    result = subprocess.run(
        ["conda", "list", "--json"], stdout=subprocess.PIPE, text=True, check=False
    )

    installed_packages = json.loads(result.stdout)

    with open(path.with_suffix(".json"), "w") as json_file:
        json.dump(installed_packages, json_file, indent=4)


@contextmanager
def all_logging_disabled(highest_level: int = logging.CRITICAL) -> Any:
    """
    A context manager that prevents any logging messages from being processed.

    Args:
        highest_level: The maximum logging level to disable. Only needs to be
            changed if a custom level greater than CRITICAL is defined.

    Example:
        ```python
        with all_logging_disabled():
            # Code here will not produce any log output
            logging.warning("This warning will not be logged")
        ```

    Reference:
        https://gist.github.com/simon-weber/7853144
    """
    previous_level = logging.root.manager.disable

    logging.disable(highest_level)

    try:
        yield
    finally:
        logging.disable(previous_level)



"""Joblib store backends for xarray objects using h5 via netcdf.

Info: Rationale
    Joblib is powerful for caching. It pickles objects into a directory. Pickling
    xarray objects however, which have their own storage solutions, seems inefficient.
    This module provides custom store backends for joblib that use h5 via netcdf for
    storage of xarray objects.
"""

import logging
import os
import warnings
from typing import Any, List

import xarray as xr
from joblib._store_backends import CacheWarning, FileSystemStoreBackend

logger = logging.getLogger(__name__)


class H5XArrayDatasetStoreBackend(FileSystemStoreBackend):
    """FileSystemStoreBackend subclass for handling xarray.Dataset objects.

    This class uses xarray's to_netcdf and open_dataset methods for Dataset objects and
    .h5 files.

    Attributes:
        location (str): The base directory for storing items.
    """

    def dump_item(self, path: List[str], item: Any, *args, **kwargs) -> None:
        """Dump an item to the store.

        If the item is an xarray.Dataset or the path ends with '.h5', use
        xarray.Dataset.to_netcdf. Otherwise, use the superclass method.

        Args:
            path: The identifier for the item in the store.
            item: The item to be stored.
            *args: Variable positional arguments passed to parent class or to_netcdf
            **kwargs: Variable keyword arguments passed to parent class or to_netcdf
        """
        is_dataset = isinstance(item, xr.Dataset)
        is_h5_file = path[-1].endswith('.h5') if path else False

        if is_dataset or is_h5_file:
            item_path = os.path.join(self.location, *path)
            nc_path = item_path if is_h5_file else os.path.join(item_path, 'output.h5')

            verbose = kwargs.get('verbose', 1)
            if verbose > 10:
                logger.info('Persisting Dataset to h5 at %s', nc_path)

            try:
                self.create_location(os.path.dirname(nc_path))
                logger.info("Store item %s", nc_path)
                # Ensure mode='w' by default but allow override through kwargs
                kwargs.setdefault('mode', 'w')
                item.to_netcdf(nc_path)
            except Exception as e:
                warnings.warn(
                    f"Unable to cache Dataset to h5. Exception: {e}.",
                    CacheWarning,
                    stacklevel=2,
                )
        else:
            super().dump_item(path, item, *args, **kwargs)

    def load_item(self, path: List[str], *args, **kwargs) -> Any:
        """Load an item from the store.

        If the path ends with '.h5' or the store contains a h5 file, use
        xarray.open_dataset. Otherwise, use the superclass method.

        Args:
            path: The identifier for the item in the store.
            *args: Variable positional arguments passed to parent class or xr.open_dataset
            **kwargs: Variable keyword arguments passed to parent class or xr.open_dataset

        Returns:
            The loaded item, either an xarray.Dataset or the original object.
        """
        item_path = os.path.join(self.location, *path)
        nc_path = (
            item_path
            if path[-1].endswith('.h5')
            else os.path.join(item_path, 'output.h5')
        )
        print(nc_path)
        if self._item_exists(nc_path):
            verbose = kwargs.get('verbose', 1)
            if verbose > 1:
                logger.info('Loading Dataset from h5 at %s', nc_path)
            try:
                return xr.open_dataset(nc_path)
            except Exception as e:
                warnings.warn(
                    f"Unable to load Dataset from h5. Exception: {e}.",
                    CacheWarning,
                    stacklevel=2,
                )
        return super().load_item(path, *args, **kwargs)

    def contains_item(self, path: List[str]) -> bool:
        """Check if there is an item at the given path.

        This method checks for both h5 and pickle files.

        Args:
            path: The identifier for the item in the store.

        Returns:
            True if the item exists in either h5 or pickle format, False otherwise.
        """
        item_path = os.path.join(self.location, *path)
        nc_filename = (
            item_path
            if path[-1].endswith('.h5')
            else os.path.join(item_path, 'output.h5')
        )
        super_filename = os.path.join(item_path, 'output.pkl')

        return self._item_exists(nc_filename) or super()._item_exists(super_filename)



import argparse
import hashlib
import io
import os
import zipfile
from typing import Dict

from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload

from flyvis import root_dir

checksums: Dict[str, str] = {
    "results_pretrained_models.zip": (
        "71c78d4070556a536b13b23ee3139cd2788aa2a9d07d430a223b4edead281db1"
    ),
    "results_umap_and_clustering.zip": (
        "6300ec1e678cde4d95300ac88e697b75fda32f6404ebe9a50f71e9f653aa9b19"
    ),
}

large_files: Dict[str, bool] = {
    "results_pretrained_models.zip": False,
    "results_umap_and_clustering.zip": False,
}


def calculate_sha256(file_path: str) -> str:
    """
    Calculate the SHA256 checksum of a file.

    Args:
        file_path: Path to the file.

    Returns:
        The SHA256 checksum as a hexadecimal string.
    """
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def download_and_unpack_files(
    folder_id: str, destination_dir: str, api_key: str, skip_large_files: bool = False
) -> None:
    """
    Download and unpack files from a Google Drive folder.

    Args:
        folder_id: The ID of the Google Drive folder.
        destination_dir: The local directory to save and unpack the files.
        api_key: The Google Drive API key.
        skip_large_files: Whether to skip downloading large files.

    Note:
        This function creates the destination directory if it doesn't exist,
        downloads ZIP files from the specified Google Drive folder, checks their
        checksums, and unpacks them.
    """
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)

    service = build("drive", "v3", developerKey=api_key)

    query = f"'{folder_id}' in parents and mimeType='application/zip'"
    results = service.files().list(q=query, fields="files(id, name)").execute()
    items = results.get("files", [])

    if not items:
        print("No files found.")
        return

    for item in items:
        file_id = item["id"]
        file_name = item["name"]

        if skip_large_files and large_files.get(file_name, False):
            print(f"Skipping {file_name}.")
            continue

        request = service.files().get_media(fileId=file_id)
        file_path = os.path.join(destination_dir, file_name)
        print(f"Downloading {file_name} to {file_path}.")

        with io.FileIO(file_path, "wb") as fh:
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            while not done:
                status, done = downloader.next_chunk()
                print(f"Progress {file_name}: {int(status.progress() * 100)}%", end="\r")
            print()

        checksum = calculate_sha256(file_path)
        if checksum != checksums[file_name]:
            print(
                f"Checksum mismatch for {file_name}. Expected {checksums[file_name]}, "
                f"got {checksum}."
            )
            return
        print(f"Checksum OK for {file_name}.")

        with zipfile.ZipFile(file_path, "r") as zip_ref:
            zip_ref.extractall(destination_dir)
        print(f"Unpacked {file_name}.")


def main() -> None:
    """
    Main function to handle command line arguments and initiate the download process.
    """
    arg_parser = argparse.ArgumentParser(
        description="Download pretrained models and UMAP clustering results. "
        "This script downloads two ZIP files from Google Drive:\n"
        "1. results_pretrained_models.zip: Contains pretrained neural networks.\n"
        "2. results_umap_and_clustering.zip: Contains UMAP embeddings and "
        "clustering.\n"
        "The files are downloaded and unpacked to the 'data' directory in the "
        "project root.",
        formatter_class=argparse.RawTextHelpFormatter,
        usage=(
            "\nflyvis download-pretrained [-h] [--skip_large_files]\n"
            "       or\n"
            "%(prog)s [-h] [--skip_large_files]\n"
        ),
    )
    arg_parser.add_argument(
        "--skip_large_files",
        action="store_true",
        help="Skip downloading large files. If set, only 'results_pretrained_models.zip' "
        "will be downloaded, and 'results_umap_and_clustering.zip' will be skipped.",
    )
    args, _ = arg_parser.parse_known_intermixed_args()

    folder_id = "15_8gPaVVJV6wGAspwkrdN8r-2NxMY4Kj"
    api_key = "AIzaSyDOy2_N7B5UjxKy5Xxeyd9WZfnDWzQ4-54"

    download_and_unpack_files(folder_id, str(root_dir), api_key, args.skip_large_files)


if __name__ == "__main__":
    main()



import numpy as np

__all__ = ["correlation", "quick_correlation_one_to_many"]


def add_noise_to_constant_rows(x, epsilon=1e-15):
    """Adds noise to rows that are constant to avoid division by zero.

    Args:
        x: m x n values

    Returns:
        m x n values
    """
    if epsilon is not None:
        x = np.array(x)
        if len(x.shape) == 1:
            x = np.expand_dims(x, axis=0)

        # check which rows are nearly constant
        is_constant = np.all(np.isclose(x, x[:, 0][:, None]), axis=1)

        # add noise to constant rows
        if is_constant.any():
            x[is_constant] += np.random.uniform(
                -epsilon, epsilon, size=x[is_constant].shape
            )
    return x


def correlation(x, x_pred, epsilon=0):
    """Slow correlation computation between one x and many x_pred values including
    significance test.
    Args:
        x: n values, array like
        x_pred: m x n values

    Returns:
        array of correlations: m values
        array of pvalues: n values
    """
    x_pred = add_noise_to_constant_rows(x_pred, epsilon=epsilon)

    from scipy.stats import pearsonr

    _correlations = [pearsonr(x, _x_pred) for _x_pred in x_pred]
    correlations = [c.statistic for c in _correlations]
    pvalues = [c.pvalue for c in _correlations]
    return np.array(correlations), np.array(pvalues)


def quick_correlation_one_to_many(x, Y, zero_nans=True):
    """Fast correlation computation between one x and many Y values.

    Args:
        x: n values, array like
        Y: m x n values

    Returns:
        array of correlations: m values
    """
    x = x - np.mean(x)
    Y = Y - np.mean(Y, axis=1, keepdims=True)
    normx = np.linalg.norm(x)
    normy = np.linalg.norm(Y, axis=1)

    if zero_nans:
        corr = (x @ Y.T) / (normx * normy)
        corr[np.isnan(corr)] = 0
        return corr
    return (x @ Y.T) / (normx * normy)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch

import flyvis
from flyvis.analysis.visualization import plots, plt_utils
from flyvis.datasets.datasets import StimulusDataset
from flyvis.datasets.sintel import AugmentedSintel
from flyvis.network.stimulus import Stimulus
from flyvis.utils import hex_utils, tensor_utils
from flyvis.utils.activity_utils import LayerActivity

__all__ = ["FindOptimalStimuli", "GenerateOptimalStimuli", "plot_stim_response"]


class FindOptimalStimuli:
    """Methods to derive optimal stimuli for cells from stimuli dataset.

    Args:
        network_view: Network view.
        stimuli: Stimuli dataset. "default" uses AugmentedSintelLum.

    Attributes:
        nv (flyvis.NetworkView): Network view.
        network (flyvis.Network): Initialized network.
        central_cells_index (list): Central cells index.
        stimuli (StimulusDataset): Stimulus dataset.
    """

    def __init__(
        self,
        network_view: flyvis.NetworkView,
        stimuli: StimulusDataset | str = "default",
    ):
        self.nv = network_view
        self.network = network_view.init_network()  # type: flyvis.Network
        for param in self.network.parameters():
            param.requires_grad = False
        self.central_cells_index = self.network.connectome.central_cells_index[:]
        self.stimuli = (
            AugmentedSintel(tasks=["lum"], dt=1 / 100, temporal_split=True)
            if stimuli == "default"
            else stimuli
        )

    def optimal_stimuli(
        self,
        cell_type: str,
        dt: float = 1 / 100,
        indices: list[int] | None = None,
    ) -> OptimalStimulus:
        """Finds optimal stimuli for a given cell type in stimuli dataset.

        Args:
            cell_type: Node type.
            dt: Time step.
            indices: Indices of stimuli.

        Returns:
            OptimalStimulus object containing the stimulus and response.
        """
        responses = self.nv.naturalistic_stimuli_responses()
        cell_responses = responses['responses'].custom.where(cell_type=cell_type)

        argmax = cell_responses.argmax(dim=("sample", "frame"))['sample'].item()
        if indices is not None:
            argmax = indices[argmax]
        nat_opt_stim = self.stimuli[argmax]["lum"]

        n_frames = nat_opt_stim.shape[0]
        initial_state = self.network.steady_state(1.0, dt, 1)
        stimulus = Stimulus(self.network.connectome, 1, n_frames)
        stimulus.zero()
        stimulus.add_input(nat_opt_stim[None])
        response = self.network(stimulus(), dt, state=initial_state).detach().cpu()
        response = LayerActivity(response, self.network.connectome, keepref=True)[
            cell_type
        ]

        return OptimalStimulus(nat_opt_stim[None, :], response[:, :, None])

    def regularized_optimal_stimuli(
        self,
        cell_type: str,
        l2_act: float = 1,
        lr: float = 1e-2,
        l2_stim: float = 1,
        n_iters: int = 100,
        dt: float = 1 / 100,
        indices: list[int] | None = None,
    ) -> RegularizedOptimalStimulus:
        """Regularizes the optimal stimulus for a given cell type.

        Maintains central node activity while minimizing mean square of input pixels.

        Args:
            cell_type: Node type.
            l2_act: L2 regularization strength for the activity.
            lr: Learning rate.
            l2_stim: L2 regularization strength for the stimulus.
            n_iters: Number of iterations.
            dt: Time step.
            indices: Indices of stimuli.

        Returns:
            RegularizedOptimalStimulus object.
        """

        optim_stimuli = self.optimal_stimuli(
            cell_type=cell_type,
            dt=dt,
            indices=indices,
        )
        non_nan = ~torch.isnan(
            optim_stimuli.stimulus[0, :, 0, optim_stimuli.stimulus.shape[-1] // 2]
        )
        reg_opt_stim = optim_stimuli.stimulus.clone()
        reg_opt_stim = reg_opt_stim[:, non_nan]
        reg_opt_stim.requires_grad = True

        central_target_response = (
            optim_stimuli.response.to(non_nan.device)[
                :, non_nan, :, optim_stimuli.response.shape[-1] // 2
            ]
            .clone()
            .detach()
            .squeeze()
        )

        optim = torch.optim.Adam([reg_opt_stim], lr=lr)

        n_frames = reg_opt_stim.shape[1]

        stim = Stimulus(self.network.connectome, 1, n_frames)

        layer_activity = LayerActivity(None, self.network.connectome, keepref=True)

        initial_state = self.network.steady_state(1.0, dt, 1)

        losses = []
        for _ in range(n_iters):
            optim.zero_grad()
            stim.zero()
            stim.add_input(reg_opt_stim)
            activities = self.network(stim(), dt, state=initial_state)
            layer_activity.update(activities)
            central_predicted_response = layer_activity.central[cell_type].squeeze()

            act_loss = (
                l2_act
                * ((central_predicted_response - central_target_response) ** 2).sum()
            )
            stim_loss = l2_stim * ((reg_opt_stim - 0.5) ** 2).mean(dim=0).sum()
            loss = act_loss + stim_loss
            loss.backward(retain_graph=True)
            optim.step()
            losses.append(loss.detach().cpu().numpy().item())

        stim.zero()
        reg_opt_stim.requires_grad = False
        stim.add_input(reg_opt_stim)
        activities = self.network(stim(), dt, state=initial_state)
        layer_activity.update(activities)

        reg_opt_stim = reg_opt_stim.detach().cpu()
        rnmei_response = layer_activity[cell_type].detach().cpu()
        central_predicted_response = central_predicted_response.detach().cpu()
        central_target_response = central_target_response.detach().cpu()
        return RegularizedOptimalStimulus(
            optim_stimuli,
            reg_opt_stim,
            rnmei_response,
            central_predicted_response,
            central_target_response,
            losses,
        )


class GenerateOptimalStimuli:
    """Methods to generate optimal stimuli for cells from random noise."""

    def __init__(self, network_view: flyvis.NetworkView):
        self.network = network_view.init_network()  # type: flyvis.Network

        for param in self.network.parameters():
            param.requires_grad = False

    def artificial_optimal_stimuli(
        self,
        cell_type: str,
        t_stim: float = 49 / 200,
        dt: float = 1 / 100,
        lr: float = 1e-2,
        weight_central: float = 1.0,
        weight_mei: float = 600 * 5,
        n_iters: int = 200,
        random_seed: int = 0,
        last_only: bool = True,
    ) -> GeneratedOptimalStimulus:
        """Generate artificial stimuli maximally exciting the central node of a type.

        Args:
            cell_type: Node type.
            t_stim: Stimulus duration.
            dt: Time step.
            lr: Learning rate.
            weight_central: Weight for central node optimization.
            weight_mei: Weight for MEI optimization.
            n_iters: Number of iterations.
            random_seed: Random seed for initialization.
            last_only: If True, optimize only the last frame.

        Returns:
            GeneratedOptimalStimulus object.
        """
        n_frames = int(t_stim / dt)
        n_hexals = hex_utils.get_num_hexals(self.network.config.connectome.extent)

        initial_state = self.network.steady_state(1.0, dt, 1)

        # Prepare pre stimulus.
        stimulus = Stimulus(self.network.connectome, 1, n_frames)

        # Initialize maximally excitatory tensors per time bin.
        torch.manual_seed(random_seed)
        art_opt_stim = torch.rand(1, n_frames, 1, n_hexals, device=flyvis.device)

        art_opt_stim.data.clamp_(0, 1)
        art_opt_stim.requires_grad = True

        # Initialize optimizer for the stimuli.
        optim = torch.optim.Adam([art_opt_stim], lr=lr)

        # Memorize surround_index for the loop.
        central_index = stimulus.central_cells_index[cell_type]

        optimize_frames = [range(n_frames)[-1]] if last_only else list(range(n_frames))

        def optimize(mei):
            # Reset stimulus.
            stimulus.zero()
            stimulus.add_input(mei)
            # Reset optimizer.
            optim.zero_grad()

            central_loss = 0
            # Stimulate network.
            activity = self.network(stimulus(), dt, state=initial_state)
            central_activity = activity[:, :, central_index].squeeze()

            central_loss -= (
                weight_central * torch.exp(central_activity[optimize_frames]).mean()
            )

            mei_loss = weight_mei * ((mei - 0.5) ** 2).mean()
            loss = central_loss + mei_loss
            loss.backward(retain_graph=True)
            optim.step()
            mei.data.clamp_(0, 1)
            return (
                loss.detach().cpu().numpy(),
                central_loss.detach().cpu().numpy(),
                mei_loss.detach().cpu().numpy(),
            )

        losses = []
        for _ in range(n_iters):
            loss = optimize(art_opt_stim)
            losses.append(loss)

        losses = np.array(losses)

        art_opt_stim.requires_grad = False
        # Stimulate network with whole sequence.
        stimulus.zero()
        stimulus.add_input(art_opt_stim)
        activity = self.network(stimulus(), dt, state=initial_state)
        responses = activity.detach().cpu().numpy()[:, :, stimulus.layer_index[cell_type]]

        art_opt_stim = art_opt_stim.cpu().numpy()
        return GeneratedOptimalStimulus(art_opt_stim, responses, losses)


@dataclass
class OptimalStimulus:
    """Optimal stimulus and response."""

    stimulus: np.ndarray
    response: np.ndarray


@dataclass
class RegularizedOptimalStimulus:
    """Regularized optimal stimulus and related data."""

    stimulus: OptimalStimulus
    regularized_stimulus: np.ndarray
    response: np.ndarray
    central_predicted_response: np.ndarray
    central_target_response: np.ndarray
    losses: np.ndarray


@dataclass
class GeneratedOptimalStimulus:
    """Generated optimal stimulus, response, and optimization losses."""

    stimulus: np.ndarray
    response: np.ndarray
    losses: np.ndarray


@dataclass
class StimResponsePlot:
    """Stimulus-response plot data and methods."""

    stim: np.ndarray
    response: np.ndarray
    dt: float
    u: np.ndarray
    v: np.ndarray
    time: np.ndarray
    t_step: np.ndarray
    t_steps_stim: np.ndarray
    t_steps_response: np.ndarray
    xmin_lattice: float
    xmax_lattice: float
    ymin_lattice: float
    ymax_lattice: float
    subtraced_baseline: bool
    steps: int
    fig: Any
    axes: Any
    time_axis: Any
    trace_axis: Any
    argmax: int
    t_argmax: float

    def __iter__(self):
        """Yield figure and axes."""
        yield from [self.fig, self.axes]

    def add_to_trace_axis(
        self,
        other: "StimResponsePlot",
        color: str | None = None,
        label: str | None = None,
        linewidth: float | None = None,
    ):
        """Add another StimResponsePlot's trace to this plot's trace axis."""
        xticks = self.trace_axis.get_xticks()
        mask = (other.time >= other.t_step.min()) & (other.time <= other.t_step.max())
        time = np.linspace(xticks.min(), xticks.max(), mask.sum())
        self.trace_axis.plot(
            time,
            other.response[mask, other.response.shape[-1] // 2],
            color=color,
            label=label,
            linewidth=linewidth,
        )


def plot_stim_response(
    stim: np.ndarray,
    response: np.ndarray,
    dt: float,
    u: np.ndarray,
    v: np.ndarray,
    max_extent: int = 6,
    subtract_baseline: bool = True,
    seconds: float = 0.2,
    steps: int = 10,
    columns: int = 10,
    suptitle: str = "",
    plot_resp: bool = True,
    hlines: bool = True,
    vlines: bool = True,
    time_axis: bool = True,
    peak_central: bool = False,
    wspace: float = -0.2,
    peak_last: bool = True,
    fontsize: int = 5,
    ylabel: str = "",
    ylabelrotation: int = 90,
    figsize: list[float] = [5, 1],
    label_peak_response: bool = False,
    fig: plt.Figure | None = None,
    axes: np.ndarray | None = None,
    crange: float | None = None,
    trace_axis: bool = False,
    trace_label: str | None = None,
    trace_axis_offset: float = 0.1,
    trace_color: str | None = None,
) -> StimResponsePlot:
    """Plot spatio-temporal stimulus and response on regular hex lattices.

    Args:
        stim: Stimulus array.
        response: Response array.
        dt: Time step.
        u: Hexagonal u-coordinates.
        v: Hexagonal v-coordinates.
        max_extent: Maximum extent of the hexagonal grid.
        subtract_baseline: Whether to subtract baseline from response.
        seconds: Duration to plot in seconds.
        steps: Number of time steps to plot.
        columns: Number of columns in the plot.
        suptitle: Super title for the plot.
        plot_resp: Whether to plot response.
        hlines: Whether to plot horizontal lines.
        vlines: Whether to plot vertical lines.
        time_axis: Whether to add a time axis.
        peak_central: Whether to center the plot around the peak.
        wspace: Width space between subplots.
        peak_last: Whether to show the peak in the last frame.
        fontsize: Font size for labels and titles.
        ylabel: Y-axis label.
        ylabelrotation: Rotation angle for y-axis label.
        figsize: Figure size.
        label_peak_response: Whether to label the peak response.
        fig: Existing figure to plot on.
        axes: Existing axes to plot on.
        crange: Color range for the plot.
        trace_axis: Whether to add a trace axis.
        trace_label: Label for the trace.
        trace_axis_offset: Offset for the trace axis.
        trace_color: Color for the trace.

    Returns:
        StimResponsePlot object containing plot data and figure.
    """
    stim = tensor_utils.to_numpy(stim).squeeze()
    mask = ~np.isnan(stim).any(axis=-1).squeeze()
    response = tensor_utils.to_numpy(response).squeeze()
    stim = stim[mask]
    response = response[mask]

    if subtract_baseline:
        response -= response[0]

    argmax = np.nanargmax(response[:, response.shape[-1] // 2])

    n_frames = response.shape[0]
    time = np.arange(n_frames) * dt
    steps = int(seconds / dt)
    t_argmax = time[argmax]

    if peak_central:
        start = argmax - steps // 2
        end = argmax + steps // 2
        if start < 0:
            start = 0
            end = steps
        peak_last = False

    if peak_last:
        start = argmax - steps
        end = argmax
        if start < 0:
            start = 0
            end = steps

    _t_steps = time[start:end]

    # resample in time in case seconds, number of columns, dt does not match
    time_index = np.linspace(0, len(_t_steps), 2 * columns, endpoint=False).astype(int)
    _t_steps = _t_steps[time_index]

    # breakpoint()
    t_steps_stim = _t_steps[0::2]
    t_steps_resp = _t_steps[1::2]

    _u, _v = hex_utils.get_hex_coords(max_extent)
    x, y = hex_utils.hex_to_pixel(_u, _v)
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()
    elev = 0
    azim = 0

    if fig is None or axes is None:
        if plot_resp:
            x, y = hex_utils.hex_rows(2, columns)
            fig, axes, pos = plt_utils.ax_scatter(
                x,
                y,
                figsize=figsize,
                hpad=0,
                wpad=0.07,
                wspace=-0.7,
                hspace=-0.5,
            )
            axes = np.array(axes).reshape(2, columns)

        else:
            fig, axes = plt_utils.divide_figure_to_grid(
                np.arange(10).reshape(1, 10),
                wspace=wspace,
                as_matrix=True,
                figsize=figsize,
            )

    crange = crange or np.abs(np.nanmax(response))
    for i, t in enumerate(t_steps_stim):
        # plot stimulus
        mask = np.where(np.abs(time - t) <= 1e-15, True, False)
        _stim = stim[mask].squeeze()
        plots.quick_hex_scatter(
            _stim,
            vmin=0,
            vmax=1,
            cbar=False,
            max_extent=max_extent,
            fig=fig,
            ax=axes[0, i],
        )

        if hlines:
            axes[0, i].hlines(elev, xmin, xmax, color="#006400", linewidth=0.25)
        if vlines:
            axes[0, i].vlines(azim, ymin, ymax, color="#006400", linewidth=0.25)

        if plot_resp:
            # --- plot response

            mask = np.where(np.abs(time - t_steps_resp[i]) <= 1e-15, True, False)
            _resp = response[mask].squeeze()
            plots.hex_scatter(
                u,
                v,
                _resp,
                fill=True,
                # edgecolor="0.3",
                # edgewidth=0.1,
                cmap=plt.cm.coolwarm,
                vmin=-crange,
                vmax=crange,
                midpoint=0,
                cbar=False,
                max_extent=max_extent,
                fig=fig,
                ax=axes[1, i],
            )
            if t_steps_resp[i] == t_argmax and label_peak_response:
                axes[1, i].set_title("peak", fontsize=fontsize)

            if hlines:
                axes[1, i].hlines(elev, xmin, xmax, color="#006400", linewidth=0.25)
            if vlines:
                axes[1, i].vlines(azim, ymin, ymax, color="#006400", linewidth=0.25)

    if trace_axis:
        left = fig.transFigure.inverted().transform(
            axes[0, 0].transData.transform((0, 0))
        )[0]
        right = fig.transFigure.inverted().transform(
            axes[-1, -1].transData.transform((0, 0))
        )[0]

        lefts, bottoms, rights, tops = np.array([
            ax.get_position().extents for ax in axes.flatten()
        ]).T

        trace_axis = fig.add_axes(
            (
                left,
                bottoms.min() - trace_axis_offset,
                right - left,
                trace_axis_offset - 0.05 * trace_axis_offset,
            ),
            label="trace_axis",
        )
        plt_utils.rm_spines(
            trace_axis, ("top", "right"), rm_yticks=False, rm_xticks=False
        )

        data_centers_in_points = np.array([
            ax.transData.transform((0, 0)) for ax in axes.flatten(order="F")
        ])
        trace_axis.tick_params(axis="both", labelsize=fontsize)
        if plot_resp:
            xticks = trace_axis.transData.inverted().transform(data_centers_in_points)[
                1::2, 0
            ]
            trace_axis.set_xticks(xticks)
            ticklabels = np.round(_t_steps * 1000, 0)
            trace_axis.set_xticklabels((ticklabels - ticklabels.max())[1::2])
        else:
            xticks = trace_axis.transData.inverted().transform(data_centers_in_points)[
                :, 0
            ]
            trace_axis.set_xticks(xticks)
            ticklabels = np.round(t_steps_stim * 1000, 0)
            trace_axis.set_xticklabels((ticklabels - ticklabels.max()))
        trace_axis.set_xlabel("time (ms)", fontsize=fontsize, labelpad=2)
        plt_utils.set_spine_tick_params(
            trace_axis,
            spinewidth=0.25,
            tickwidth=0.25,
            ticklength=3,
            ticklabelpad=2,
            spines=("top", "right", "bottom", "left"),
        )
        xlim = trace_axis.get_xlim()
        mask = (time >= _t_steps.min()) & (time <= _t_steps.max())

        time = np.linspace(xticks.min(), xticks.max(), mask.sum())

        trace_axis.plot(
            time,
            response[mask, response.shape[-1] // 2],
            label=trace_label,
            color=trace_color,
        )
        trace_axis.set_xlim(*xlim)
        trace_axis.set_ylabel("central\nresponse", fontsize=fontsize)
        # flyvis.plots.trim_axis(trace_axis)

        time_axis = False

    if time_axis:
        left = fig.transFigure.inverted().transform(
            axes[0, 0].transData.transform((0, 0))
        )[0]
        right = fig.transFigure.inverted().transform(
            axes[-1, -1].transData.transform((0, 0))
        )[0]

        lefts, bottoms, rights, tops = np.array([
            ax.get_position().extents for ax in axes.flatten()
        ]).T

        time_axis = fig.add_axes((left, bottoms.min(), right - left, 0.01))
        plt_utils.rm_spines(
            time_axis, ("left", "top", "right"), rm_yticks=True, rm_xticks=False
        )

        data_centers_in_points = np.array([
            ax.transData.transform((0, 0)) for ax in axes.flatten(order="F")
        ])
        time_axis.tick_params(axis="both", labelsize=fontsize)
        if plot_resp:
            time_axis.set_xticks(
                time_axis.transData.inverted().transform(data_centers_in_points)[1::2, 0]
            )
            ticklabels = np.round(_t_steps * 1000, 0)
            time_axis.set_xticklabels((ticklabels - ticklabels.max())[1::2])
        else:
            time_axis.set_xticks(
                time_axis.transData.inverted().transform(data_centers_in_points)[:, 0]
            )
            ticklabels = np.round(t_steps_stim * 1000, 0)
            time_axis.set_xticklabels((ticklabels - ticklabels.max()))
        time_axis.set_xlabel("time (ms)", fontsize=fontsize, labelpad=2)
        plt_utils.set_spine_tick_params(
            time_axis,
            spinewidth=0.25,
            tickwidth=0.25,
            ticklength=3,
            ticklabelpad=2,
            spines=("top", "right", "bottom", "left"),
        )

    if ylabel:
        lefts, bottoms, rights, tops = np.array([
            ax.get_position().extents for ax in axes.flatten()
        ]).T
        ylabel_axis = fig.add_axes((
            lefts.min(),
            bottoms.min(),
            0.01,
            tops.max() - bottoms.min(),
        ))
        plt_utils.rm_spines(
            ylabel_axis,
            ("left", "top", "right", "bottom"),
            rm_yticks=True,
            rm_xticks=True,
        )
        ylabel_axis.set_ylabel(ylabel, fontsize=fontsize, rotation=ylabelrotation)
        ylabel_axis.patch.set_alpha(0)

    if plot_resp and ylabel is not None:
        axes[0, 0].annotate(
            "stimulus",
            xy=(0, 0.5),
            ha="right",
            va="center",
            fontsize=fontsize,
            rotation=90,
            xycoords="axes fraction",
        )
        axes[1, 0].annotate(
            "response",
            xy=(0, 0.5),
            ha="right",
            va="center",
            fontsize=fontsize,
            rotation=90,
            xycoords="axes fraction",
        )

    if suptitle:
        lefts, bottoms, rights, tops = np.array([
            ax.get_position().extents for ax in axes.flatten()
        ]).T
        fig.suptitle(suptitle, fontsize=fontsize, y=tops.max(), va="bottom")

    plt_utils.set_spine_tick_params(
        fig.axes[-1],
        spinewidth=0.25,
        tickwidth=0.25,
        ticklength=3,
        ticklabelpad=2,
        spines=("top", "right", "bottom", "left"),
    )

    fig.crange = crange

    return StimResponsePlot(
        stim,
        response,
        dt,
        u,
        v,
        time,
        _t_steps,
        t_steps_stim,
        t_steps_resp,
        xmin,
        xmax,
        ymin,
        ymax,
        subtract_baseline,
        steps,
        fig,
        axes,
        time_axis,
        trace_axis,
        argmax,
        t_argmax,
    )



from dataclasses import dataclass

import numpy as np
import pytest
import torch
from datamate import Namespace

import flyvis
from flyvis import Network
from flyvis.connectome.connectome import init_connectome, register_connectome
from flyvis.network.network import IntegrationWarning
from flyvis.utils.tensor_utils import AutoDeref


@pytest.fixture(scope="module")
def network() -> Network:
    network = Network(
        connectome=Namespace(
            type="ConnectomeFromAvgFilters",
            file="fib25-fib19_v2.2.json",
            extent=15,
            n_syn_fill=1,
        ),
        dynamics=Namespace(type="PPNeuronIGRSynapses", activation=Namespace(type="relu")),
        node_config=Namespace(
            bias=Namespace(
                type="RestingPotential",
                groupby=["type"],
                initial_dist="Normal",
                mode="sample",
                requires_grad=True,
                mean=0.5,
                std=0.05,
                penalize=Namespace(activity=True),
                seed=0,
            ),
            time_const=Namespace(
                type="TimeConstant",
                groupby=["type"],
                initial_dist="Value",
                value=0.05,
                requires_grad=True,
            ),
        ),
        edge_config=Namespace(
            sign=Namespace(
                type="SynapseSign",
                initial_dist="Value",
                requires_grad=False,
                groupby=["source_type", "target_type"],
            ),
            syn_count=Namespace(
                type="SynapseCount",
                initial_dist="Lognormal",
                mode="mean",
                requires_grad=False,
                std=1.0,
                groupby=["source_type", "target_type", "du", "dv"],
            ),
            syn_strength=Namespace(
                type="SynapseCountScaling",
                initial_dist="Value",
                requires_grad=True,
                scale=0.01,
                clamp="non_negative",
                groupby=["source_type", "target_type"],
            ),
        ),
    )
    return network


def test_init(network):
    assert isinstance(network, Network)
    assert hasattr(network, "connectome")
    assert hasattr(network, "dynamics")
    assert hasattr(network, "node_params")
    assert hasattr(network, "edge_params")
    assert hasattr(network, "config")
    assert hasattr(network, "symmetry_config")
    assert hasattr(network, "clamp_config")
    assert hasattr(network, "_state_hooks")
    assert hasattr(network, "num_parameters")
    assert hasattr(network, "stimulus")


def test_param_api(network):
    param_api = network._param_api()
    assert len(param_api) == 4
    assert param_api.get("nodes", None) is not None
    assert param_api.get("edges", None) is not None
    assert param_api.get("sources", None) is not None
    assert param_api.get("targets", None) is not None
    assert isinstance(
        param_api.get("nodes").get(list(param_api.get("nodes").keys())[0], None),
        flyvis.utils.tensor_utils.RefTensor,
    )
    assert len(param_api.sources.bias) == network.n_edges
    assert len(param_api.targets.bias) == network.n_edges


def test_target_sum(network):
    x = torch.ones(2, network.n_edges)
    y = network.target_sum(x)
    assert y.shape == (2, network.n_nodes)


def test_initial_state(network):
    param_api = network._param_api()
    state = network._initial_state(param_api, 2)
    assert state.get("nodes", None) is not None
    assert state.get("edges", None) is not None
    assert state.get("sources", None) is not None
    assert state.get("targets", None) is not None
    assert state.nodes.activity.shape == (2, network.n_nodes)


def test_next_state(network):
    param_api = network._param_api()
    state = network._initial_state(param_api, 2)
    x_t = torch.ones(2, network.n_nodes)
    dt = 0.02
    state = network._next_state(param_api, state, x_t, dt)
    assert state.get("nodes", None) is not None
    assert state.get("edges", None) is not None
    assert state.get("sources", None) is not None
    assert state.get("targets", None) is not None
    assert state.sources.activity.shape == (2, network.n_edges)


def test_state_hook(network):
    assert not network._state_hooks
    assert not hasattr(network, "times_hooked")

    param_api = network._param_api()

    def record_hook(state):
        assert state.get("nodes", None) is not None
        assert state.get("edges", None) is not None
        assert state.nodes.activity.shape == (2, network.n_nodes)
        if not hasattr(network, "times_hooked"):
            network.times_hooked = 0
        network.times_hooked += 1

    network.register_state_hook(record_hook)
    state = network._initial_state(param_api, 2)
    assert network.times_hooked == 1
    x_t = torch.ones(2, network.n_nodes)
    dt = 0.02
    state = network._next_state(param_api, state, x_t, dt)
    assert network.times_hooked == 2

    def modify_state_hook(state, value):
        activity = state.nodes.activity.clone()
        activity *= value
        state.nodes.update(activity=activity)
        return state

    _activity = state.nodes.activity.clone()
    network.register_state_hook(modify_state_hook, value=2)
    modified_state = network._state_api(state)
    assert (modified_state.nodes.activity == 2 * _activity).all()
    assert network.times_hooked == 3


def test_forward(network):
    x = torch.ones(2, 20, network.n_nodes).random_(2)
    activity = network.forward(x, 1 / 50)
    assert activity.shape == (2, 20, network.n_nodes)
    assert activity.grad_fn.name() == "StackBackward0"


def test_simulate(network):
    x = torch.ones(2, 20, 1, 721).random_(2)
    activity = network.simulate(x, 1 / 50)
    assert activity.shape == (2, 20, network.n_nodes)

    with pytest.raises(ValueError):
        network.simulate(torch.ones(20, 1, 721).random_(2), 1 / 50)

    with pytest.warns(IntegrationWarning):
        network.simulate(x, 1 / 49)


def test_steady_state(network: Network):
    steady_state = network.steady_state(1, 1 / 20, 2, 0.5, None, False)

    assert isinstance(steady_state, AutoDeref)

    assert list(steady_state.keys()) == ["nodes", "edges", "sources", "targets"]

    assert steady_state["nodes"]["activity"].shape == (2, network.n_nodes)

    assert steady_state["sources"]["activity"].shape == (2, network.n_edges)

    assert steady_state["targets"]["activity"].shape == (2, network.n_edges)


def test_fade_in_state(network):
    initial_frames = torch.ones(2, 1, 721).uniform_()

    steady_state = network.fade_in_state(1, 1 / 20, initial_frames, None, False)

    assert isinstance(steady_state, AutoDeref)
    assert list(steady_state.keys()) == ["nodes", "edges", "sources", "targets"]
    assert steady_state["nodes"]["activity"].shape == (2, network.n_nodes)
    assert steady_state["sources"]["activity"].shape == (2, network.n_edges)
    assert steady_state["targets"]["activity"].shape == (2, network.n_edges)


@register_connectome
class DiagonalConnectome:
    @dataclass
    class nodes:
        index = np.arange(100)

    @dataclass
    class edges:
        source_index = np.arange(100)
        target_index = np.arange(100)


def test_connectome_protocol():
    connectome = init_connectome(type="DiagonalConnectome")
    assert isinstance(connectome, DiagonalConnectome)

    network = Network(
        connectome=Namespace(type="DiagonalConnectome"),
        dynamics=Namespace(),
        node_config=Namespace(),
        edge_config=Namespace(),
        stimulus_config=Namespace(),
    )
    assert isinstance(network, Network)



from copy import deepcopy
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
import torch
from datamate import Directory, Namespace

from flyvis import results_dir
from flyvis.network.ensemble import Ensemble, TaskError
from flyvis.network.ensemble_view import EnsembleView
from flyvis.network.network import IntegrationWarning, Network

pytestmark = pytest.mark.require_download


@pytest.fixture(scope="function")
def ensemble() -> Ensemble:
    models = [results_dir / f"flow/0000/{i:03}" for i in range(4)]
    ensemble = Ensemble(
        models,
        best_checkpoint_fn_kwargs={
            "validation_subdir": "validation",
            "loss_file_name": "loss",
        },
    )
    return ensemble


def test_ensemble_init(ensemble):
    assert isinstance(ensemble, Ensemble)
    assert len(ensemble) == 4
    assert hasattr(ensemble, "names")
    assert hasattr(ensemble, "name")
    assert hasattr(ensemble, "path")
    assert hasattr(ensemble, "model_paths")
    assert hasattr(ensemble, "dir")


def test_yield_networks(ensemble):
    networks = list(ensemble.yield_networks())
    assert len(networks) == len(ensemble)
    assert isinstance(networks[0], Network)


def test_simulate(ensemble: Ensemble):
    x = torch.ones(1, 2, 1, 721).random_(2)
    network = next(ensemble.yield_networks())
    with pytest.warns(IntegrationWarning):
        activity = np.array(list(ensemble.simulate(x, 1)))
    assert activity.shape == (len(ensemble), 1, 2, network.n_nodes)

    with pytest.raises(ValueError):
        activity = np.array(list(ensemble.simulate(torch.ones(1, 2, 721).random_(2), 1)))


def test_validation_losses(ensemble):
    losses = ensemble.validation_losses()
    assert len(losses) == len(ensemble)


def test_rank_by_validation_error(ensemble):
    sorted_names = deepcopy(ensemble.names)

    # destroy current task sorting in place
    np.random.shuffle(ensemble.names)

    random_names = deepcopy(ensemble.names)

    with ensemble.rank_by_validation_error():
        assert ensemble.names == sorted_names
        assert ensemble.names != random_names

    with ensemble.rank_by_validation_error(reverse=True):
        assert ensemble.names == sorted_names[::-1]

    assert ensemble.names == random_names
    ensemble.names = sorted_names


def test_ratio(ensemble):
    names = deepcopy(ensemble.names)

    with ensemble.ratio(best=0.5):
        assert len(ensemble) == int(0.5 * len(names))
        assert ensemble.names == names[: int(0.5 * len(names))]

    with ensemble.ratio(worst=0.5):
        assert len(ensemble) == int(0.5 * len(names))
        assert ensemble.names == names[int(0.5 * len(names)) :]

    with ensemble.ratio(best=1 / 3, worst=1 / 3):
        assert len(ensemble) == int(2 / 3 * len(names))
        assert ensemble.names == [
            *names[: int(len(names) * 1 / 3)],
            *names[len(names) - int(len(names) * 1 / 3) :],
        ]

    assert ensemble.names == names


def test_task_error(ensemble):
    task_error = ensemble.task_error()
    assert isinstance(task_error, TaskError)
    assert len(task_error.values) == len(ensemble)
    assert len(task_error.values) == len(task_error.colors)


@pytest.mark.slow
def test_cluster_indices():
    ensemble = Ensemble(
        "flow/0000",
        best_checkpoint_fn_kwargs={
            "validation_subdir": "validation",
            "loss_file_name": "loss",
        },
    )
    for cell_type in ensemble.connectome.unique_cell_types[:].astype(str):
        cluster_indices = ensemble.cluster_indices(cell_type)
        assert isinstance(cluster_indices, dict)
        assert type(cluster_indices[0]) is np.ndarray
        assert cluster_indices[0].dtype is np.dtype("int64")
        assert set(
            sorted(np.concatenate(list(cluster_indices.values()), axis=0))
        ).issubset(set(np.arange(len(ensemble)).tolist()))


def test_loss_histogram(ensemble: Ensemble):
    ensemble_view = EnsembleView(
        ensemble,
        best_checkpoint_fn_kwargs={
            "validation_subdir": "validation",
            "loss_file_name": "loss",
        },
    )
    fig, ax = ensemble_view.task_error_histogram()
    fig.show()
    plt.close(fig)


@pytest.mark.slow
def test_responses():
    ensemble = Ensemble(
        results_dir / "flow/0000",
        best_checkpoint_fn_kwargs={
            "validation_subdir": "validation",
            "loss_file_name": "loss",
        },
    )

    test_data_dir = Directory(Path(__file__).parent / "data")
    assert "responses" in test_data_dir
    test_responses = test_data_dir.responses[:]
    assert test_responses.shape == (50, 12, 1, 65)
    config = test_data_dir.config
    assert config == Namespace(
        method="steady_state",
        t_pre=0.25,
        dt=1 / 50,
        batch_size=1,
        value=0.5,
        state=None,
        grad=False,
        return_last=False,
    )
    method = config.pop("method")
    ensemble_responses = []
    central_cells_index = None
    for network in ensemble.yield_networks():
        if central_cells_index is None:
            central_cells_index = network.connectome.central_cells_index[:]
        if isinstance(method, str):
            method = getattr(network, method)
        states = method(**config)
        activity = np.array([s.nodes.activity.cpu().numpy() for s in states])[
            :, :, central_cells_index
        ]
        ensemble_responses.append(activity)
    ensemble_responses = np.array(ensemble_responses)
    assert ensemble_responses.shape == test_responses.shape
    assert np.allclose(ensemble_responses, test_responses, atol=1e-4)



"""Script to run synthetic recordings for each model of an ensemble."""

import argparse
import logging
from importlib import resources
from typing import List

from flyvis.utils.compute_cloud_utils import launch_range

logging.basicConfig(
    format="[%(asctime)s] [%(filename)s:%(lineno)d] %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


def run_synthetic_recordings(args: argparse.Namespace, kwargs: List[str]) -> None:
    """
    Launch synthetic recording job for each model of an ensemble.

    Args:
        args: Command-line arguments.
        kwargs: Additional keyword arguments as a list of strings.
    """
    launch_range(
        args.start,
        args.end,
        args.ensemble_id,
        args.task_name,
        args.nP,
        args.gpu,
        args.q,
        args.synthetic_recordings_script,
        args.dry,
        kwargs,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "Run synthetic recordings for each model of an ensemble on the compute cloud."
        ),
        usage=(
            "\nflyvis record [-h] [...] --ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "[options] [synthetic_recordings_script_args...]\n"
            "       or\n"
            "%(prog)s [-h] [...] --ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "[options] [synthetic_recordings_script_args...]\n"
            "\n"
            "For a full list of options and default arguments, run: "
            "flyvis synthetic-recordings-single --help"
        ),
    )
    parser.add_argument("--start", type=int, default=0, help="Start id of ensemble.")
    parser.add_argument("--end", type=int, default=50, help="End id of ensemble.")
    parser.add_argument("--nP", type=int, default=4, help="Number of processors.")
    parser.add_argument("--gpu", type=str, default="num=1", help="Number of GPUs.")
    parser.add_argument("--q", type=str, default="gpu_l4", help="Queue.")
    parser.add_argument(
        "--ensemble_id",
        type=int,
        required=True,
        help="Id of the ensemble, e.g. 0045.",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        required=True,
        help="Name given to the task, e.g., flow.",
    )
    DEFAULT_SCRIPT = str(
        resources.files("flyvis_cli") / "analysis" / "synthetic_recordings_single.py"
    )
    parser.add_argument(
        "--synthetic_recordings_script",
        type=str,
        default=DEFAULT_SCRIPT,
        help=f"Script to run for synthetic recordings. Default: {DEFAULT_SCRIPT}",
    )
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Perform a dry run without actually launching jobs.",
    )

    args, kwargs = parser.parse_known_intermixed_args()
    run_synthetic_recordings(args, kwargs)



"""Methods to access and process network activity and currents based on cell types.

Example:
    Get the activity of all cells of a certain cell type
    (note, `LayerActivity` is not the network split into feedforward layers in the
    machine learning sense but the activity of all cells by cell type):
    ```python
    layer_activity = LayerActivity(activity, network.connectome)
    T4a_response = layer_activity.T4a
    T5a_response = layer_activity.T5a
    T4b_central_response = layer_activity.central.T4a
    ```
"""

import weakref
from textwrap import wrap
from typing import List, Union

import numpy as np
import torch
from numpy.typing import NDArray

from flyvis.connectome import ConnectomeFromAvgFilters, ReceptiveFields
from flyvis.utils import nodes_edges_utils

__all__ = [
    "CentralActivity",
    "LayerActivity",
    "SourceCurrentView",
]


class CellTypeActivity(dict):
    """Base class for attribute-style access to network activity based on cell types.

    Args:
        keepref: Whether to keep a reference to the activity. This may not be desired
            during training to avoid memory issues.

    Attributes:
        activity: Weak reference to the activity.
        keepref: Whether to keep a reference to the activity.
        unique_cell_types: List of unique cell types.
        input_indices: Indices of input cells.
        output_indices: Indices of output cells.

    Note:
        Activity is stored as a weakref by default for memory efficiency
        during training. Set keepref=True to keep a reference for analysis.
    """

    def __init__(self, keepref: bool = False):
        self.keepref = keepref
        self.activity: Union[weakref.ref, NDArray, torch.Tensor] = None
        self.unique_cell_types: List[str] = []
        self.input_indices: NDArray = np.array([])
        self.output_indices: NDArray = np.array([])

    def __dir__(self) -> List[str]:
        return list(set([*dict.__dir__(self), *dict.__iter__(self)]))

    def __len__(self) -> int:
        return len(self.unique_cell_types)

    def __iter__(self):
        yield from self.unique_cell_types

    def __repr__(self) -> str:
        return "Activity of: \n{}".format("\n".join(wrap(", ".join(list(self)))))

    def update(self, activity: Union[NDArray, torch.Tensor]) -> None:
        """Update the activity reference."""
        self.activity = activity

    def _slices(self, n: int) -> tuple:
        return tuple(slice(None) for _ in range(n))

    def __getattr__(self, key):
        activity = self.activity() if not self.keepref else self.activity
        if activity is None:
            return
        if isinstance(key, list):
            index = np.stack(list(map(lambda key: dict.__getitem__(self, key), key)))
            slices = self._slices(len(activity.shape) - 1)
            slices += (index,)
            return activity[slices]
        elif key == slice(None):
            return activity
        elif key in self.unique_cell_types:
            slices = self._slices(len(activity.shape) - 1)
            slices += (dict.__getitem__(self, key),)
            return activity[slices]
        elif key == "output":
            slices = self._slices(len(activity.shape) - 1)
            slices += (self.output_indices,)
            return activity[slices]
        elif key == "input":
            slices = self._slices(len(activity.shape) - 1)
            slices += (self.input_indices,)
            return activity[slices]
        elif key in self.__dict__:
            return self.__dict__[key]
        else:
            raise ValueError(f"{key}")

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setattr__(self, key, value):
        if key == "activity" and value is not None:
            if self.keepref is False:
                value = weakref.ref(value)
            object.__setattr__(self, key, value)
        else:
            object.__setattr__(self, key, value)


class CentralActivity(CellTypeActivity):
    """Attribute-style access to central cell activity of a cell type.

    Args:
        activity: Activity of shape (..., n_cells).
        connectome: Connectome directory with reference to required attributes.
        keepref: Whether to keep a reference to the activity.

    Attributes:
        activity: Activity of shape (..., n_cells).
        unique_cell_types: Array of unique cell types.
        index: NodeIndexer instance.
        input_indices: Array of input indices.
        output_indices: Array of output indices.
    """

    def __init__(
        self,
        activity: Union[NDArray, torch.Tensor],
        connectome: ConnectomeFromAvgFilters,
        keepref: bool = False,
    ):
        super().__init__(keepref)
        self.index = nodes_edges_utils.NodeIndexer(connectome)

        unique_cell_types = connectome.unique_cell_types[:]
        input_cell_types = connectome.input_cell_types[:]
        output_cell_types = connectome.output_cell_types[:]
        self.input_indices = np.array([
            np.nonzero(unique_cell_types == t)[0] for t in input_cell_types
        ])
        self.output_indices = np.array([
            np.nonzero(unique_cell_types == t)[0] for t in output_cell_types
        ])
        self.activity = activity
        self.unique_cell_types = unique_cell_types.astype(str)

    def __getattr__(self, key):
        activity = self.activity() if not self.keepref else self.activity
        if activity is None:
            return
        if isinstance(key, list):
            index = np.stack(list(map(lambda key: self.index[key], key)))
            slices = self._slices(len(activity.shape) - 1)
            slices += (index,)
            return activity[slices]
        elif key == slice(None):
            return activity
        elif key in self.index.unique_cell_types:
            slices = self._slices(len(activity.shape) - 1)
            slices += (self.index[key],)
            return activity[slices]
        elif key == "output":
            slices = self._slices(len(activity.shape) - 1)
            slices += (self.output_indices,)
            return activity[slices]
        elif key == "input":
            slices = self._slices(len(activity.shape) - 1)
            slices += (self.input_indices,)
            return activity[slices]
        elif key in self.__dict__:
            return self.__dict__[key]
        else:
            raise ValueError(f"{key}")

    def __setattr__(self, key, value):
        if key == "activity" and value is not None:
            if len(self.index.unique_cell_types) != value.shape[-1]:
                slices = self._slices(len(value.shape) - 1)
                slices += (self.index.central_cells_index,)
                value = value[slices]
                self.keepref = True
            if self.keepref is False:
                value = weakref.ref(value)
            object.__setattr__(self, key, value)
        else:
            object.__setattr__(self, key, value)

    def __len__(self):
        return len(self.unique_cell_types)

    def __iter__(self):
        for cell_type in self.unique_cell_types:
            yield cell_type


class LayerActivity(CellTypeActivity):
    """Attribute-style access to hex-lattice activity (cell-type specific).

    Args:
        activity: Activity of shape (..., n_cells).
        connectome: Connectome directory with reference to required attributes.
        keepref: Whether to keep a reference to the activity.
        use_central: Whether to use central activity.

    Attributes:
        central: CentralActivity instance for central nodes.
        activity: Activity of shape (..., n_cells).
        connectome: Connectome directory.
        unique_cell_types: Array of unique cell types.
        input_indices: Array of input indices.
        output_indices: Array of output indices.
        input_cell_types: Array of input cell types.
        output_cell_types: Array of output cell types.
        n_nodes: Number of nodes.

    Note:
        The name `LayerActivity` might change in future as it is misleading.
        This is not a feedforward layer in the machine learning sense but the
        activity of all cells of a certain cell-type.

    Example:

        Central activity can be accessed by:
        ```python
        a = LayerActivity(activity, network.connectome)
        central_T4a = a.central.T4a
        ```

        Also allows 'virtual types' that are the sum of individuals:
        ```python
        a = LayerActivity(activity, network.connectome)
        summed_a = a['L2+L4']
        ```
    """

    def __init__(
        self,
        activity: Union[NDArray, torch.Tensor],
        connectome: ConnectomeFromAvgFilters,
        keepref: bool = False,
        use_central: bool = True,
    ):
        super().__init__(keepref)
        self.keepref = keepref

        self.use_central = use_central
        if use_central:
            self.central = CentralActivity(activity, connectome, keepref)

        self.activity = activity
        self.connectome = connectome
        self.unique_cell_types = connectome.unique_cell_types[:].astype("str")
        for cell_type in self.unique_cell_types:
            index = connectome.nodes.layer_index[cell_type][:]
            self[cell_type] = index

        _cell_types = self.connectome.nodes.type[:]
        self.input_indices = np.array([
            np.nonzero(_cell_types == t)[0] for t in self.connectome.input_cell_types
        ])
        self.output_indices = np.array([
            np.nonzero(_cell_types == t)[0] for t in self.connectome.output_cell_types
        ])
        self.input_cell_types = self.connectome.input_cell_types[:].astype(str)
        self.output_cell_types = self.connectome.output_cell_types[:].astype(str)
        self.n_nodes = len(self.connectome.nodes.type)

    def __setattr__(self, key, value):
        if key == "activity" and value is not None:
            if self.keepref is False:
                value = weakref.ref(value)

            if self.use_central:
                self.central.__setattr__(key, value)

            object.__setattr__(self, key, value)
        else:
            object.__setattr__(self, key, value)


class SourceCurrentView:
    """Create views of source currents for a target type.

    Args:
        rfs: ReceptiveFields instance.
        currents: Current values.

    Attributes:
        target_type: Target cell type.
        source_types: List of source cell types.
        rfs: ReceptiveFields instance.
        currents: Current values.
    """

    def __init__(self, rfs: ReceptiveFields, currents: Union[NDArray, torch.Tensor]):
        self.target_type = rfs.target_type
        self.source_types = list(rfs)
        self.rfs = rfs
        self.currents = currents

    def __getattr__(self, key: str) -> Union[NDArray, torch.Tensor]:
        if key in self.source_types:
            return np.take(self.currents, self.rfs[key].index, axis=-1)
        return object.__getattr__(self, key)

    def __getitem__(self, key: str) -> Union[NDArray, torch.Tensor]:
        return self.__getattr__(key)

    def update(self, currents: Union[NDArray, torch.Tensor]) -> None:
        """Update the currents."""
        self.currents = currents
